#!/usr/bin/python
# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: online_server_info
short_description: Gather information about Online servers
description:
  - Gather information about the servers.
  - U(https://www.online.net/en/dedicated-server).
author:
  - "Remy Leone (@remyleone)"
extends_documentation_fragment:
  - community.general._online
  - community.general._attributes
  - community.general._attributes.info_module
"""

EXAMPLES = r"""
- name: Gather Online server information
  community.general.online_server_info:
    api_token: '0d1627e8-bbf0-44c5-a46f-5c4d3aef033f'
  register: result

- ansible.builtin.debug:
    msg: "{{ result.online_server_info }}"
"""

RETURN = r"""
online_server_info:
  description:
    - Response from Online API.
    - 'For more details please refer to: U(https://console.online.net/en/api/).'
  returned: success
  type: list
  elements: dict
  sample:
    [
      {
        "abuse": "abuse@example.com",
        "anti_ddos": false,
        "bmc": {
          "session_key": null
        },
        "boot_mode": "normal",
        "contacts": {
          "owner": "foobar",
          "tech": "foobar"
        },
        "disks": [
          {
            "$ref": "/api/v1/server/hardware/disk/68452"
          },
          {
            "$ref": "/api/v1/server/hardware/disk/68453"
          }
        ],
        "drive_arrays": [
          {
            "disks": [
              {
                "$ref": "/api/v1/server/hardware/disk/68452"
              },
              {
                "$ref": "/api/v1/server/hardware/disk/68453"
              }
            ],
            "raid_controller": {
              "$ref": "/api/v1/server/hardware/raidController/9910"
            },
            "raid_level": "RAID1"
          }
        ],
        "hardware_watch": true,
        "hostname": "sd-42",
        "id": 42,
        "ip": [
          {
            "address": "195.154.172.149",
            "mac": "28:92:4a:33:5e:c6",
            "reverse": "195-154-172-149.rev.poneytelecom.eu.",
            "switch_port_state": "up",
            "type": "public"
          },
          {
            "address": "10.90.53.212",
            "mac": "28:92:4a:33:5e:c7",
            "reverse": null,
            "switch_port_state": "up",
            "type": "private"
          }
        ],
        "last_reboot": "2018-08-23T08:32:03.000Z",
        "location": {
          "block": "A",
          "datacenter": "DC3",
          "position": 19,
          "rack": "A23",
          "room": "4 4-4"
        },
        "network": {
          "ip": [
            "195.154.172.149"
          ],
          "ipfo": [],
          "private": [
            "10.90.53.212"
          ]
        },
        "offer": "Pro-1-S-SATA",
        "os": {
          "name": "FreeBSD",
          "version": "11.1-RELEASE"
        },
        "power": "ON",
        "proactive_monitoring": false,
        "raid_controllers": [
          {
            "$ref": "/api/v1/server/hardware/raidController/9910"
          }
        ],
        "support": "Basic service level"
      }
    ]
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._online import (
    Online,
    OnlineException,
    online_argument_spec,
)


class OnlineServerInfo(Online):
    def __init__(self, module):
        super().__init__(module)
        self.name = "api/v1/server"

    def _get_server_detail(self, server_path):
        try:
            return self.get(path=server_path).json
        except OnlineException as exc:
            self.module.fail_json(msg=f"A problem occurred while fetching: {server_path} ({exc})")

    def all_detailed_servers(self):
        servers_api_path = self.get_resources()

        server_data = (self._get_server_detail(server_api_path) for server_api_path in servers_api_path)

        return [s for s in server_data if s is not None]


def main():
    module = AnsibleModule(
        argument_spec=online_argument_spec(),
        supports_check_mode=True,
    )

    try:
        servers_info = OnlineServerInfo(module).all_detailed_servers()
        module.exit_json(online_server_info=servers_info)
    except OnlineException as exc:
        module.fail_json(msg=exc.message)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2021, Phillipe Smith <phsmithcc@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: rundeck_job_run
short_description: Run a Rundeck job
description:
  - This module runs a Rundeck job specified by ID.
author: "Phillipe Smith (@phsmith)"
version_added: 3.8.0
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  job_id:
    type: str
    description:
      - The job unique ID.
    required: true
  job_options:
    type: dict
    description:
      - The job options for the steps.
      - Numeric values must be quoted.
  filter_nodes:
    type: str
    description:
      - Filter the nodes where the jobs must run.
      - See U(https://docs.rundeck.com/docs/manual/11-node-filters.html#node-filter-syntax).
  run_at_time:
    type: str
    description:
      - Schedule the job execution to run at specific date and time.
      - ISO-8601 date and time format like V(2021-10-05T15:45:00-03:00).
  loglevel:
    type: str
    description:
      - Log level configuration.
    choices: [debug, verbose, info, warn, error]
    default: info
  wait_execution:
    type: bool
    description:
      - Wait until the job finished the execution.
    default: true
  wait_execution_delay:
    type: int
    description:
      - Delay, in seconds, between job execution status check requests.
    default: 5
  wait_execution_timeout:
    type: int
    description:
      - Job execution wait timeout in seconds.
      - If the timeout is reached, the job is aborted.
      - Keep in mind that there is a sleep based on O(wait_execution_delay) after each job status check.
    default: 120
  abort_on_timeout:
    type: bool
    description:
      - Send a job abort request if exceeded the O(wait_execution_timeout) specified.
    default: false
extends_documentation_fragment:
  - community.general._rundeck
  - ansible.builtin.url
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Run a Rundeck job
  community.general.rundeck_job_run:
    url: "https://rundeck.example.org"
    api_version: 39
    api_token: "mytoken"
    job_id: "xxxxxxxxxxxxxxxxx"
  register: rundeck_job_run

- name: Show execution info
  ansible.builtin.debug:
    var: rundeck_job_run.execution_info

- name: Run a Rundeck job with options
  community.general.rundeck_job_run:
    url: "https://rundeck.example.org"
    api_version: 39
    api_token: "mytoken"
    job_id: "xxxxxxxxxxxxxxxxx"
    job_options:
      option_1: "value_1"
      option_2: "value_3"
      option_3: "value_3"
  register: rundeck_job_run

- name: Run a Rundeck job with timeout, delay between status check and abort on timeout
  community.general.rundeck_job_run:
    url: "https://rundeck.example.org"
    api_version: 39
    api_token: "mytoken"
    job_id: "xxxxxxxxxxxxxxxxx"
    wait_execution_timeout: 30
    wait_execution_delay: 10
    abort_on_timeout: true
  register: rundeck_job_run

- name: Schedule a Rundeck job
  community.general.rundeck_job_run:
    url: "https://rundeck.example.org"
    api_version: 39
    api_token: "mytoken"
    job_id: "xxxxxxxxxxxxxxxxx"
    run_at_time: "2021-10-05T15:45:00-03:00"
  register: rundeck_job_schedule

- name: Fire-and-forget a Rundeck job
  community.general.rundeck_job_run:
    url: "https://rundeck.example.org"
    api_version: 39
    api_token: "mytoken"
    job_id: "xxxxxxxxxxxxxxxxx"
    wait_execution: false
  register: rundeck_job_run
"""

RETURN = r"""
execution_info:
  description: Rundeck job execution metadata.
  returned: always
  type: dict
  sample:
    {
      "msg": "Job execution succeeded!",
      "execution_info": {
        "id": 1,
        "href": "https://rundeck.example.org/api/39/execution/1",
        "permalink": "https://rundeck.example.org/project/myproject/execution/show/1",
        "status": "succeeded",
        "project": "myproject",
        "executionType": "user",
        "user": "admin",
        "date-started": {
          "unixtime": 1633449020784,
          "date": "2021-10-05T15:50:20Z"
        },
        "date-ended": {
          "unixtime": 1633449026358,
          "date": "2021-10-05T15:50:26Z"
        },
        "job": {
          "id": "697af0c4-72d3-4c15-86a3-b5bfe3c6cb6a",
          "averageDuration": 4917,
          "name": "Test",
          "group": "",
          "project": "myproject",
          "description": "",
          "options": {
            "exit_code": "0"
          },
          "href": "https://rundeck.example.org/api/39/job/697af0c4-72d3-4c15-86a3-b5bfe3c6cb6a",
          "permalink": "https://rundeck.example.org/project/myproject/job/show/697af0c4-72d3-4c15-86a3-b5bfe3c6cb6a"
        },
        "description": "sleep 5 && echo 'Test!' && exit ${option.exit_code}",
        "argstring": "-exit_code 0",
        "serverUUID": "5b9a1438-fa3a-457e-b254-8f3d70338068",
        "successfulNodes": [
          "localhost"
        ],
        "output": "Test!"
      }
    }
"""

# Modules import
from datetime import datetime, timedelta
from time import sleep
from urllib.parse import quote

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._rundeck import api_argument_spec, api_request


class RundeckJobRun:
    def __init__(self, module):
        self.module = module
        self.url = self.module.params["url"]
        self.api_version = self.module.params["api_version"]
        self.job_id = self.module.params["job_id"]
        self.job_options = self.module.params["job_options"] or {}
        self.filter_nodes = self.module.params["filter_nodes"] or ""
        self.run_at_time = self.module.params["run_at_time"] or ""
        self.loglevel = self.module.params["loglevel"].upper()
        self.wait_execution = self.module.params["wait_execution"]
        self.wait_execution_delay = self.module.params["wait_execution_delay"]
        self.wait_execution_timeout = self.module.params["wait_execution_timeout"]
        self.abort_on_timeout = self.module.params["abort_on_timeout"]

        for k, v in self.job_options.items():
            if not isinstance(v, str):
                self.module.exit_json(msg=f"Job option '{k}' value must be a string", execution_info={})

    def job_status_check(self, execution_id):
        response = dict()
        timeout = False
        due = datetime.now() + timedelta(seconds=self.wait_execution_timeout)

        while not timeout:
            endpoint = f"execution/{execution_id}"
            response = api_request(module=self.module, endpoint=endpoint)[0]
            output = api_request(module=self.module, endpoint=f"execution/{execution_id}/output")
            log_output = "\n".join([x["log"] for x in output[0]["entries"]])
            response.update({"output": log_output})

            if response["status"] == "aborted":
                break
            elif response["status"] == "scheduled":
                self.module.exit_json(
                    msg=f"Job scheduled to run at {self.run_at_time}", execution_info=response, changed=True
                )
            elif response["status"] == "failed":
                self.module.fail_json(msg="Job execution failed", execution_info=response)
            elif response["status"] == "succeeded":
                self.module.exit_json(msg="Job execution succeeded!", execution_info=response)

            if datetime.now() >= due:
                timeout = True
                break

            # Wait for 5s before continue
            sleep(self.wait_execution_delay)

        response.update({"timed_out": timeout})
        return response

    def job_run(self):
        response, info = api_request(
            module=self.module,
            endpoint=f"job/{quote(self.job_id)}/run",
            method="POST",
            data={
                "loglevel": self.loglevel,
                "options": self.job_options,
                "runAtTime": self.run_at_time,
                "filter": self.filter_nodes,
            },
        )

        if info["status"] != 200:
            self.module.fail_json(msg=info["msg"])

        if not self.wait_execution:
            self.module.exit_json(msg="Job run send successfully!", execution_info=response)

        job_status = self.job_status_check(response["id"])

        if job_status["timed_out"]:
            if self.abort_on_timeout:
                api_request(module=self.module, endpoint=f"execution/{response['id']}/abort", method="GET")

                abort_status = self.job_status_check(response["id"])

                self.module.fail_json(
                    msg="Job execution aborted due the timeout specified", execution_info=abort_status
                )

            self.module.fail_json(msg="Job execution timed out", execution_info=job_status)


def main():
    argument_spec = api_argument_spec()
    argument_spec.update(
        dict(
            job_id=dict(required=True, type="str"),
            job_options=dict(type="dict"),
            filter_nodes=dict(type="str"),
            run_at_time=dict(type="str"),
            wait_execution=dict(type="bool", default=True),
            wait_execution_delay=dict(type="int", default=5),
            wait_execution_timeout=dict(type="int", default=120),
            abort_on_timeout=dict(type="bool", default=False),
            loglevel=dict(type="str", choices=["debug", "verbose", "info", "warn", "error"], default="info"),
        )
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=False)

    if module.params["api_version"] < 14:
        module.fail_json(msg="API version should be at least 14")

    rundeck = RundeckJobRun(module)
    rundeck.job_run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2017, Milan Ilic <milani@nordeus.com>
# Copyright (c) 2019, Jan Meerkamp <meerkamp@dvv.de>
# Copyright (c) 2025, Tom Paine <github@aioue.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
module: one_vm
short_description: Creates or terminates OpenNebula instances
description:
  - Manages OpenNebula instances.
requirements:
  - pyone
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  api_url:
    description:
      - URL of the OpenNebula RPC server.
      - It is recommended to use HTTPS so that the username/password are not transferred over the network unencrypted.
      - If not set then the value of the E(ONE_URL) environment variable is used.
    type: str
  api_username:
    description:
      - Name of the user to login into the OpenNebula RPC server. If not set then the value of the E(ONE_USERNAME) environment
        variable is used.
    type: str
  api_password:
    description:
      - Password of the user to login into OpenNebula RPC server. If not set then the value of the E(ONE_PASSWORD) environment
        variable is used. if both O(api_username) or O(api_password) are not set, then it tries to authenticate with ONE auth
        file. Default path is C(~/.one/one_auth).
      - Set environment variable E(ONE_AUTH) to override this path.
    type: str
  template_name:
    description:
      - Name of VM template to use to create a new instance.
    type: str
  template_id:
    description:
      - ID of a VM template to use to create a new instance.
    type: int
  vm_start_on_hold:
    description:
      - Set to true to put VM on hold while creating.
    default: false
    type: bool
  instance_ids:
    description:
      - 'A list of instance IDs used for states: V(absent), V(running), V(rebooted), V(poweredoff).'
    aliases: ['ids']
    type: list
    elements: int
  state:
    description:
      - V(present) - create instances from a template specified with C(template_id)/C(template_name).
      - V(running) - run instances.
      - V(poweredoff) - power-off instances.
      - V(rebooted) - reboot instances.
      - V(absent) - terminate instances.
    choices: ["present", "absent", "running", "rebooted", "poweredoff"]
    default: present
    type: str
  hard:
    description:
      - Reboot, power-off or terminate instances C(hard).
    default: false
    type: bool
  wait:
    description:
      - Wait for the instance to reach its desired state before returning. Keep in mind if you are waiting for instance to
        be in running state it does not mean that you are able to SSH on that machine only that boot process have started
        on that instance. See the example using the M(ansible.builtin.wait_for) module for details.
    default: true
    type: bool
  wait_timeout:
    description:
      - How long before wait gives up, in seconds.
    default: 300
    type: int
  attributes:
    description:
      - A dictionary of key/value attributes to add to new instances, or for setting C(state) of instances with these attributes.
      - Keys are case insensitive and OpenNebula automatically converts them to upper case.
      - Be aware V(NAME) is a special attribute which sets the name of the VM when it is deployed.
      - C(#) character(s) can be appended to the C(NAME) and the module automatically adds indexes to the names of VMs.
      - 'For example: V(NAME: foo-###) would create VMs with names V(foo-000), V(foo-001),...'
      - When used with O(count_attributes) and O(exact_count) the module matches the base name without the index part.
    default: {}
    type: dict
  labels:
    description:
      - A list of labels to associate with new instances, or for setting C(state) of instances with these labels.
    default: []
    type: list
    elements: str
  count_attributes:
    description:
      - A dictionary of key/value attributes that can only be used with O(exact_count) to determine how many nodes based on
        a specific attributes criteria should be deployed. This can be expressed in multiple ways and is shown in the EXAMPLES
        section.
    type: dict
  count_labels:
    description:
      - A list of labels that can only be used with O(exact_count) to determine how many nodes based on a specific labels
        criteria should be deployed. This can be expressed in multiple ways and is shown in the EXAMPLES section.
    type: list
    elements: str
  count:
    description:
      - Number of instances to launch.
    default: 1
    type: int
  exact_count:
    description:
      - Indicates how many instances that match O(count_attributes) and O(count_labels) parameters should be deployed. Instances
        are either created or terminated based on this value.
      - B(NOTE:) Instances with the least IDs are terminated first.
    type: int
  mode:
    description:
      - Set permission mode of the instance in octet format, for example V(0600) to give owner C(use) and C(manage) and nothing
        to group and others.
    type: str
  owner_id:
    description:
      - ID of the user which is set as the owner of the instance.
    type: int
  group_id:
    description:
      - ID of the group which is set as the group of the instance.
    type: int
  memory:
    description:
      - The size of the memory for new instances (in MB, GB, ..).
    type: str
  disk_size:
    description:
      - The size of the disk created for new instances (in MB, GB, TB,...).
      - B(NOTE:) If The Template hats Multiple Disks the Order of the Sizes is matched against the order specified in O(template_id)/O(template_name).
    type: list
    elements: str
  cpu:
    description:
      - Percentage of CPU divided by 100 required for the new instance. Half a processor is written 0.5.
    type: float
  vcpu:
    description:
      - Number of CPUs (cores) the new VM uses.
    type: int
  networks:
    description:
      - A list of dictionaries with network parameters. See examples for more details.
    default: []
    type: list
    elements: dict
  disk_saveas:
    description:
      - Creates an image from a VM disk.
      - It is a dictionary where you have to specify C(name) of the new image.
      - Optionally you can specify C(disk_id) of the disk you want to save. By default C(disk_id) is 0.
      - B(NOTE:) This operation is only performed on the first VM (if more than one VM ID is passed) and the VM has to be
        in the C(poweredoff) state.
      - Also this operation fails if an image with specified C(name) already exists.
    type: dict
  persistent:
    description:
      - Create a private persistent copy of the template plus any image defined in DISK, and instantiate that copy.
    default: false
    type: bool
    version_added: '0.2.0'
  datastore_id:
    description:
      - Name of Datastore to use to create a new instance.
    version_added: '0.2.0'
    type: int
  datastore_name:
    description:
      - Name of Datastore to use to create a new instance.
    version_added: '0.2.0'
    type: str
  updateconf:
    description:
      - When O(instance_ids) is provided, updates running VMs with the C(updateconf) API call.
      - When new VMs are being created, emulates the C(updateconf) API call using direct template merge.
      - Allows for complete modifications of the C(CONTEXT) attribute.
      - 'Supported attributes include:'
      - B(BACKUP_CONFIG:) V(BACKUP_VOLATILE), V(FS_FREEZE), V(INCREMENT_MODE), V(KEEP_LAST), V(MODE);
      - B(CONTEXT:) (Any value, except V(ETH*). Variable substitutions are made);
      - B(CPU_MODEL:) V(FEATURES), V(MODEL);
      - B(FEATURES:) V(ACPI), V(APIC), V(GUEST_AGENT), V(HYPERV), V(IOTHREADS), V(LOCALTIME), V(PAE), V(VIRTIO_BLK_QUEUES),
        V(VIRTIO_SCSI_QUEUES);
      - B(GRAPHICS:) V(COMMAND), V(KEYMAP), V(LISTEN), V(PASSWD), V(PORT), V(TYPE);
      - B(INPUT:) V(BUS), V(TYPE);
      - B(OS:) V(ARCH), V(BOOT), V(BOOTLOADER), V(FIRMWARE), V(INITRD), V(KERNEL), V(KERNEL_CMD), V(MACHINE), V(ROOT), V(SD_DISK_BUS),
        V(UUID);
      - B(RAW:) V(DATA), V(DATA_VMX), V(TYPE), V(VALIDATE);
      - B(VIDEO:) V(ATS), V(IOMMU), V(RESOLUTION), V(TYPE), V(VRAM).
    type: dict
    version_added: 6.3.0
author:
  - "Milan Ilic (@ilicmilan)"
  - "Jan Meerkamp (@meerkampdvv)"
"""


EXAMPLES = r"""
- name: Create a new instance
  community.general.one_vm:
    template_id: 90
  register: result

- name: Print VM properties
  ansible.builtin.debug:
    msg: result

- name: Deploy a new VM on hold
  community.general.one_vm:
    template_name: 'app1_template'
    vm_start_on_hold: 'True'

- name: Deploy a new VM and set its name to 'foo'
  community.general.one_vm:
    template_name: 'app1_template'
    attributes:
      name: foo

- name: Deploy a new VM and set its group_id and mode
  community.general.one_vm:
    template_id: 90
    group_id: 16
    mode: 660

- name: Deploy a new VM  as persistent
  community.general.one_vm:
    template_id: 90
    persistent: true

- name: Change VM's permissions to 640
  community.general.one_vm:
    instance_ids: 5
    mode: 640

- name: Deploy 2 new instances and set memory, vcpu, disk_size and 3 networks
  community.general.one_vm:
    template_id: 15
    disk_size: 35.2 GB
    memory: 4 GB
    vcpu: 4
    count: 2
    networks:
      - NETWORK_ID: 27
      - NETWORK: "default-network"
        NETWORK_UNAME: "app-user"
        SECURITY_GROUPS: "120,124"
      - NETWORK_ID: 27
        SECURITY_GROUPS: "10"

- name: Deploy a new instance which uses a Template with two Disks
  community.general.one_vm:
    template_id: 42
    disk_size:
      - 35.2 GB
      - 50 GB
    memory: 4 GB
    vcpu: 4
    count: 1
    networks:
      - NETWORK_ID: 27

- name: "Deploy an new instance with attribute 'bar: bar1' and set its name to 'foo'"
  community.general.one_vm:
    template_id: 53
    attributes:
      name: foo
      bar: bar1

- name: "Enforce that 2 instances with attributes 'foo1: app1' and 'foo2: app2' are deployed"
  community.general.one_vm:
    template_id: 53
    attributes:
      foo1: app1
      foo2: app2
    exact_count: 2
    count_attributes:
      foo1: app1
      foo2: app2

- name: Enforce that 4 instances with an attribute 'bar' are deployed
  community.general.one_vm:
    template_id: 53
    attributes:
      name: app
      bar: bar2
    exact_count: 4
    count_attributes:
      bar:

# Deploy 2 new instances with attribute 'foo: bar' and labels 'app1' and 'app2' and names in format 'fooapp-##'
# Names will be: fooapp-00 and fooapp-01
- name: Deploy 2 new instances
  community.general.one_vm:
    template_id: 53
    attributes:
      name: fooapp-##
      foo: bar
    labels:
      - app1
      - app2
    count: 2

# Deploy 2 new instances with attribute 'app: app1' and names in format 'fooapp-###'
# Names will be: fooapp-002 and fooapp-003
- name: Deploy 2 new instances
  community.general.one_vm:
    template_id: 53
    attributes:
      name: fooapp-###
      app: app1
    count: 2

# Reboot all instances with name in format 'fooapp-#'
# Instances 'fooapp-00', 'fooapp-01', 'fooapp-002' and 'fooapp-003' will be rebooted
- name: Reboot all instances with names in a certain format
  community.general.one_vm:
    attributes:
      name: fooapp-#
    state: rebooted

# Enforce that only 1 instance with name in format 'fooapp-#' is deployed
# The task will delete oldest instances, so only the 'fooapp-003' will remain
- name: Enforce that only 1 instance with name in a certain format is deployed
  community.general.one_vm:
    template_id: 53
    exact_count: 1
    count_attributes:
      name: fooapp-#

- name: Deploy an new instance with a network
  community.general.one_vm:
    template_id: 53
    networks:
      - NETWORK_ID: 27
  register: vm

- name: Wait for SSH to come up
  ansible.builtin.wait_for:
    port: 22
    host: '{{ vm.instances[0].networks[0].ip }}'

- name: Terminate VMs by ids
  community.general.one_vm:
    instance_ids:
      - 153
      - 160
    state: absent

- name: Reboot all VMs that have labels 'foo' and 'app1'
  community.general.one_vm:
    labels:
      - foo
      - app1
    state: rebooted

- name: "Fetch all VMs that have name 'foo' and attribute 'app: bar'"
  community.general.one_vm:
    attributes:
      name: foo
      app: bar
  register: results

- name: Deploy 2 new instances with labels 'foo1' and 'foo2'
  community.general.one_vm:
    template_name: app_template
    labels:
      - foo1
      - foo2
    count: 2

- name: Enforce that only 1 instance with label 'foo1' will be running
  community.general.one_vm:
    template_name: app_template
    labels:
      - foo1
    exact_count: 1
    count_labels:
      - foo1

- name: Terminate all instances that have attribute foo
  community.general.one_vm:
    template_id: 53
    exact_count: 0
    count_attributes:
      foo:

- name: "Power-off the VM and save VM's disk with id=0 to the image with name 'foo-image'"
  community.general.one_vm:
    instance_ids: 351
    state: poweredoff
    disk_saveas:
      name: foo-image

- name: "Save VM's disk with id=1 to the image with name 'bar-image'"
  community.general.one_vm:
    instance_ids: 351
    disk_saveas:
      name: bar-image
      disk_id: 1

- name: "Deploy 2 new instances with a custom 'start script'"
  community.general.one_vm:
    template_name: app_template
    count: 2
    updateconf:
      CONTEXT:
        START_SCRIPT: ip r r 169.254.16.86/32 dev eth0

- name: "Add a custom 'start script' to a running VM"
  community.general.one_vm:
    instance_ids: 351
    updateconf:
      CONTEXT:
        START_SCRIPT: ip r r 169.254.16.86/32 dev eth0

- name: "Update SSH public keys inside the VM's context"
  community.general.one_vm:
    instance_ids: 351
    updateconf:
      CONTEXT:
        SSH_PUBLIC_KEY: |-
          ssh-rsa ...
          ssh-ed25519 ...
"""

RETURN = r"""
instances_ids:
  description: A list of instances IDs whose state is changed or which are fetched with O(instance_ids) option.
  type: list
  returned: success
  sample: [1234, 1235]
instances:
  description: A list of instances info whose state is changed or which are fetched with O(instance_ids) option.
  type: list
  elements: dict
  returned: success
  contains:
    vm_id:
      description: VM ID.
      type: int
      sample: 153
    vm_name:
      description: VM name.
      type: str
      sample: foo
    template_id:
      description: VM's template ID.
      type: int
      sample: 153
    group_id:
      description: VM's group ID.
      type: int
      sample: 1
    group_name:
      description: VM's group name.
      type: str
      sample: one-users
    owner_id:
      description: VM's owner ID.
      type: int
      sample: 143
    owner_name:
      description: VM's owner name.
      type: str
      sample: app-user
    mode:
      description: VM's mode.
      type: str
      returned: success
      sample: 660
    state:
      description: State of an instance.
      type: str
      sample: ACTIVE
    lcm_state:
      description: Lcm state of an instance that is only relevant when the state is ACTIVE.
      type: str
      sample: RUNNING
    cpu:
      description: Percentage of CPU divided by 100.
      type: float
      sample: 0.2
    vcpu:
      description: Number of CPUs (cores).
      type: int
      sample: 2
    memory:
      description: The size of the memory in MB.
      type: str
      sample: 4096 MB
    disk_size:
      description: The size of the disk in MB.
      type: str
      sample: 20480 MB
    networks:
      description: A list of dictionaries with info about IP, NAME, MAC, SECURITY_GROUPS for each NIC.
      type: list
      sample:
        [
          {
            "ip": "10.120.5.33",
            "mac": "02:00:0a:78:05:21",
            "name": "default-test-private",
            "security_groups": "0,10"
          },
          {
            "ip": "10.120.5.34",
            "mac": "02:00:0a:78:05:22",
            "name": "default-test-private",
            "security_groups": "0"
          }
        ]
    uptime_h:
      description: Uptime of the instance in hours.
      type: int
      sample: 35
    labels:
      description: A list of string labels that are associated with the instance.
      type: list
      sample: ["foo", "spec-label"]
    attributes:
      description: A dictionary of key/values attributes that are associated with the instance.
      type: dict
      sample:
        {
          "HYPERVISOR": "kvm",
          "LOGO": "images/logos/centos.png",
          "TE_GALAXY": "bar",
          "USER_INPUTS": null
        }
    updateconf:
      description: A dictionary of key/values attributes that are set with the updateconf API call.
      type: dict
      version_added: 6.3.0
      sample:
        {
          "OS": {
            "ARCH": "x86_64"
          },
          "CONTEXT": {
            "START_SCRIPT": "ip r r 169.254.16.86/32 dev eth0",
            "SSH_PUBLIC_KEY": "ssh-rsa ...\\nssh-ed25519 ..."
          }
        }
tagged_instances:
  description:
    - A list of instances info based on a specific attributes and/or labels that are specified with O(count_attributes) and
      O(count_labels) options.
  type: list
  elements: dict
  returned: success
  contains:
    vm_id:
      description: VM ID.
      type: int
      sample: 153
    vm_name:
      description: VM name.
      type: str
      sample: foo
    template_id:
      description: VM's template ID.
      type: int
      sample: 153
    group_id:
      description: VM's group ID.
      type: int
      sample: 1
    group_name:
      description: VM's group name.
      type: str
      sample: one-users
    owner_id:
      description: VM's user ID.
      type: int
      sample: 143
    owner_name:
      description: VM's user name.
      type: str
      sample: app-user
    mode:
      description: VM's mode.
      type: str
      returned: success
      sample: 660
    state:
      description: State of an instance.
      type: str
      sample: ACTIVE
    lcm_state:
      description: Lcm state of an instance that is only relevant when the state is ACTIVE.
      type: str
      sample: RUNNING
    cpu:
      description: Percentage of CPU divided by 100.
      type: float
      sample: 0.2
    vcpu:
      description: Number of CPUs (cores).
      type: int
      sample: 2
    memory:
      description: The size of the memory in MB.
      type: str
      sample: 4096 MB
    disk_size:
      description: The size of the disk in MB.
      type: list
      sample: ["20480 MB", "10240 MB"]
    networks:
      description: A list of dictionaries with info about IP, NAME, MAC, SECURITY_GROUPS for each NIC.
      type: list
      sample:
        [
          {
            "ip": "10.120.5.33",
            "mac": "02:00:0a:78:05:21",
            "name": "default-test-private",
            "security_groups": "0,10"
          },
          {
            "ip": "10.120.5.34",
            "mac": "02:00:0a:78:05:22",
            "name": "default-test-private",
            "security_groups": "0"
          }
        ]
    uptime_h:
      description: Uptime of the instance in hours.
      type: int
      sample: 35
    labels:
      description: A list of string labels that are associated with the instance.
      type: list
      sample: ["foo", "spec-label"]
    attributes:
      description: A dictionary of key/values attributes that are associated with the instance.
      type: dict
      sample:
        {
          "HYPERVISOR": "kvm",
          "LOGO": "images/logos/centos.png",
          "TE_GALAXY": "bar",
          "USER_INPUTS": null
        }
    updateconf:
      description: A dictionary of key/values attributes that are set with the updateconf API call.
      type: dict
      version_added: 6.3.0
      sample:
        {
          "OS": {
            "ARCH": "x86_64"
          },
          "CONTEXT": {
            "START_SCRIPT": "ip r r 169.254.16.86/32 dev eth0",
            "SSH_PUBLIC_KEY": "ssh-rsa ...\\nssh-ed25519 ..."
          }
        }
"""

try:
    import pyone

    HAS_PYONE = True
except ImportError:
    HAS_PYONE = False

import copy
import os
import re
import time
from dataclasses import dataclass

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.dict_transformations import dict_merge

from ansible_collections.community.general.plugins.module_utils._opennebula import flatten, render


@dataclass
class AuthParams:
    url: str
    username: str
    password: str


# Updateconf attributes documentation: https://docs.opennebula.io/6.10/integration_and_development/system_interfaces/api.html#one-vm-updateconf
UPDATECONF_ATTRIBUTES = {
    "OS": ["ARCH", "MACHINE", "KERNEL", "INITRD", "BOOTLOADER", "BOOT", "SD_DISK_BUS", "UUID", "FIRMWARE"],
    "CPU_MODEL": ["MODEL", "FEATURES"],
    "FEATURES": [
        "ACPI",
        "PAE",
        "APIC",
        "LOCALTIME",
        "HYPERV",
        "GUEST_AGENT",
        "VIRTIO_BLK_QUEUES",
        "VIRTIO_SCSI_QUEUES",
        "IOTHREADS",
    ],
    "INPUT": ["TYPE", "BUS"],
    "GRAPHICS": ["TYPE", "LISTEN", "PORT", "PASSWD", "KEYMAP", "COMMAND"],
    "VIDEO": ["ATS", "IOMMU", "RESOLUTION", "TYPE", "VRAM"],
    "RAW": ["DATA", "DATA_VMX", "TYPE", "VALIDATE"],
    "CONTEXT": [],
    "BACKUP_CONFIG": ["FS_FREEZE", "KEEP_LAST", "BACKUP_VOLATILE", "MODE", "INCREMENT_MODE"],
}


def check_updateconf(module, to_check):
    """Checks if attributes are compatible with one.vm.updateconf API call."""
    for attr, subattributes in to_check.items():
        if attr not in UPDATECONF_ATTRIBUTES:
            module.fail_json(msg=f"'{attr}' is not a valid VM attribute.")
        if not UPDATECONF_ATTRIBUTES[attr]:
            continue
        for subattr in subattributes:
            if subattr not in UPDATECONF_ATTRIBUTES[attr]:
                module.fail_json(msg=f"'{subattr}' is not a valid VM subattribute of '{attr}'")


def parse_updateconf(vm_template):
    """Extracts 'updateconf' attributes from a VM template."""
    updateconf = {}
    for attr, subattributes in vm_template.items():
        if attr not in UPDATECONF_ATTRIBUTES:
            continue
        tmp = {}
        for subattr, value in subattributes.items():
            if UPDATECONF_ATTRIBUTES[attr] and subattr not in UPDATECONF_ATTRIBUTES[attr]:
                continue
            tmp[subattr] = value
        if tmp:
            updateconf[attr] = tmp
    return updateconf


def get_template(module, client, predicate):
    pool = client.templatepool.info(-2, -1, -1, -1)
    # Filter -2 means fetch all templates user can Use
    found = 0
    found_template = None
    template_name = ""

    for template in pool.VMTEMPLATE:
        if predicate(template):
            found = found + 1
            found_template = template
            template_name = template.NAME

    if found == 0:
        return None
    elif found > 1:
        module.fail_json(msg=f"There are more templates with name: {template_name}")
    return found_template


def get_template_by_name(module, client, template_name):
    return get_template(module, client, lambda template: template_name == template.NAME)


def get_template_by_id(module, client, template_id):
    return get_template(module, client, lambda template: template_id == template.ID)


def get_template_id(module, client, requested_id, requested_name):
    template = (
        get_template_by_id(module, client, requested_id)
        if requested_id is not None
        else get_template_by_name(module, client, requested_name)
    )
    if template:
        return template.ID
    else:
        return None


def get_datastore(module, client, predicate):
    pool = client.datastorepool.info()
    found = 0
    found_datastore = None
    datastore_name = ""

    for datastore in pool.DATASTORE:
        if predicate(datastore):
            found = found + 1
            found_datastore = datastore
            datastore_name = datastore.NAME

    if found == 0:
        return None
    elif found > 1:
        module.fail_json(msg=f"There are more datastores with name: {datastore_name}")
    return found_datastore


def get_datastore_by_name(module, client, datastore_name):
    return get_datastore(module, client, lambda datastore: datastore_name == datastore.NAME)


def get_datastore_by_id(module, client, datastore_id):
    return get_datastore(module, client, lambda datastore: datastore_id == datastore.ID)


def get_datastore_id(module, client, requested_id, requested_name):
    datastore = (
        get_datastore_by_id(module, client, requested_id)
        if requested_id
        else get_datastore_by_name(module, client, requested_name)
    )
    if datastore:
        return datastore.ID
    else:
        return None


def get_vm_by_id(client, vm_id):
    try:
        vm = client.vm.info(int(vm_id))
    except BaseException:
        return None
    return vm


def get_vms_by_ids(module, client, state, ids):
    vms = []

    for vm_id in ids:
        vm = get_vm_by_id(client, vm_id)
        if vm is None and state != "absent":
            module.fail_json(msg=f"There is no VM with id={vm_id}")
        vms.append(vm)

    return vms


def get_vm_info(client, vm):
    vm = client.vm.info(vm.ID)

    networks_info = []

    disk_size = []
    if "DISK" in vm.TEMPLATE:
        if isinstance(vm.TEMPLATE["DISK"], list):
            for disk in vm.TEMPLATE["DISK"]:
                disk_size.append(f"{disk['SIZE']} MB")
        else:
            disk_size.append(f"{vm.TEMPLATE['DISK']['SIZE']} MB")

    if "NIC" in vm.TEMPLATE:
        if isinstance(vm.TEMPLATE["NIC"], list):
            for nic in vm.TEMPLATE["NIC"]:
                networks_info.append(
                    {
                        "ip": nic.get("IP", ""),
                        "mac": nic.get("MAC", ""),
                        "name": nic.get("NETWORK", ""),
                        "security_groups": nic.get("SECURITY_GROUPS", ""),
                    }
                )
        else:
            networks_info.append(
                {
                    "ip": vm.TEMPLATE["NIC"].get("IP", ""),
                    "mac": vm.TEMPLATE["NIC"].get("MAC", ""),
                    "name": vm.TEMPLATE["NIC"].get("NETWORK", ""),
                    "security_groups": vm.TEMPLATE["NIC"].get("SECURITY_GROUPS", ""),
                }
            )

    current_time = time.localtime()
    vm_start_time = time.localtime(vm.STIME)

    vm_uptime = time.mktime(current_time) - time.mktime(vm_start_time)
    vm_uptime /= 60 * 60

    permissions_str = parse_vm_permissions(client, vm)

    # LCM_STATE is VM's sub-state that is relevant only when STATE is ACTIVE
    vm_lcm_state = None
    if VM_STATES.index("ACTIVE") == vm.STATE:
        vm_lcm_state = LCM_STATES[vm.LCM_STATE]

    vm_labels, vm_attributes = get_vm_labels_and_attributes_dict(client, vm.ID)

    updateconf = parse_updateconf(vm.TEMPLATE)

    info = {
        "template_id": int(vm.TEMPLATE["TEMPLATE_ID"]),
        "vm_id": vm.ID,
        "vm_name": vm.NAME,
        "state": VM_STATES[vm.STATE],
        "lcm_state": vm_lcm_state,
        "owner_name": vm.UNAME,
        "owner_id": vm.UID,
        "networks": networks_info,
        "disk_size": disk_size,
        "memory": f"{vm.TEMPLATE['MEMORY']} MB",
        "vcpu": vm.TEMPLATE["VCPU"],
        "cpu": vm.TEMPLATE["CPU"],
        "group_name": vm.GNAME,
        "group_id": vm.GID,
        "uptime_h": int(vm_uptime),
        "attributes": vm_attributes,
        "mode": permissions_str,
        "labels": vm_labels,
        "updateconf": updateconf,
    }

    return info


def parse_vm_permissions(client, vm):
    vm_PERMISSIONS = client.vm.info(vm.ID).PERMISSIONS

    owner_octal = int(vm_PERMISSIONS.OWNER_U) * 4 + int(vm_PERMISSIONS.OWNER_M) * 2 + int(vm_PERMISSIONS.OWNER_A)
    group_octal = int(vm_PERMISSIONS.GROUP_U) * 4 + int(vm_PERMISSIONS.GROUP_M) * 2 + int(vm_PERMISSIONS.GROUP_A)
    other_octal = int(vm_PERMISSIONS.OTHER_U) * 4 + int(vm_PERMISSIONS.OTHER_M) * 2 + int(vm_PERMISSIONS.OTHER_A)

    permissions = str(owner_octal) + str(group_octal) + str(other_octal)

    return permissions


def set_vm_permissions(module, client, vms, permissions):
    changed = False

    for vm in vms:
        vm = client.vm.info(vm.ID)
        old_permissions = parse_vm_permissions(client, vm)
        changed = changed or old_permissions != permissions

        if not module.check_mode and old_permissions != permissions:
            permissions_str = bin(int(permissions, base=8))[2:]  # 600 -> 110000000
            mode_bits = [int(d) for d in permissions_str]
            try:
                client.vm.chmod(
                    vm.ID,
                    mode_bits[0],
                    mode_bits[1],
                    mode_bits[2],
                    mode_bits[3],
                    mode_bits[4],
                    mode_bits[5],
                    mode_bits[6],
                    mode_bits[7],
                    mode_bits[8],
                )
            except pyone.OneAuthorizationException:
                module.fail_json(
                    msg="Permissions changing is unsuccessful, but instances are present if you deployed them."
                )

    return changed


def set_vm_ownership(module, client, vms, owner_id, group_id):
    changed = False

    for vm in vms:
        vm = client.vm.info(vm.ID)
        if owner_id is None:
            owner_id = vm.UID
        if group_id is None:
            group_id = vm.GID

        changed = changed or owner_id != vm.UID or group_id != vm.GID

        if not module.check_mode and (owner_id != vm.UID or group_id != vm.GID):
            try:
                client.vm.chown(vm.ID, owner_id, group_id)
            except pyone.OneAuthorizationException:
                module.fail_json(
                    msg="Ownership changing is unsuccessful, but instances are present if you deployed them."
                )

    return changed


def update_vm(module, client, vm, updateconf_dict):
    changed = False
    if not updateconf_dict:
        return changed

    before = client.vm.info(vm.ID).TEMPLATE

    client.vm.updateconf(vm.ID, render(updateconf_dict), 1)  # 1: Merge new template with the existing one.

    after = client.vm.info(vm.ID).TEMPLATE

    changed = before != after
    return changed


def update_vms(module, client, vms, *args):
    changed = False
    for vm in vms:
        changed = update_vm(module, client, vm, *args) or changed
    return changed


def get_size_in_MB(module, size_str):
    SYMBOLS = ["B", "KB", "MB", "GB", "TB"]

    s = size_str
    init = size_str
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == ".":
        num += s[0]
        s = s[1:]
    num = float(num)
    symbol = s.strip()

    if symbol not in SYMBOLS:
        module.fail_json(msg=f"Cannot interpret {init!r} {symbol!r} {num}")

    prefix = {"B": 1}

    for i, s in enumerate(SYMBOLS[1:]):
        prefix[s] = 1 << (i + 1) * 10

    size_in_bytes = int(num * prefix[symbol])
    size_in_MB = size_in_bytes / (1024 * 1024)

    return size_in_MB


def create_vm(
    module,
    client,
    template_id,
    attributes_dict,
    labels_list,
    disk_size,
    network_attrs_list,
    vm_start_on_hold,
    vm_persistent,
    updateconf_dict,
):
    if attributes_dict:
        vm_name = attributes_dict.get("NAME", "")

    template = client.template.info(template_id).TEMPLATE

    disk_count = len(flatten(template.get("DISK", [])))
    if disk_size:
        size_count = len(flatten(disk_size))
        # check if the number of disks is correct
        if disk_count != size_count:
            module.fail_json(msg=f"This template has {disk_count} disks but you defined {size_count}")

    vm_extra_template = dict_merge(template or {}, attributes_dict or {})
    vm_extra_template = dict_merge(
        vm_extra_template,
        {
            "LABELS": ",".join(labels_list),
            "NIC": flatten(network_attrs_list, extract=True),
            "DISK": flatten(
                [
                    disk
                    if not size
                    else dict_merge(
                        disk,
                        {
                            "SIZE": str(int(get_size_in_MB(module, size))),
                        },
                    )
                    for disk, size in zip(
                        flatten(template.get("DISK", [])),
                        flatten(disk_size or [None] * disk_count),
                    )
                    if disk is not None
                ],
                extract=True,
            ),
        },
    )
    vm_extra_template = dict_merge(vm_extra_template, updateconf_dict or {})

    try:
        vm_id = client.template.instantiate(
            template_id, vm_name, vm_start_on_hold, render(vm_extra_template), vm_persistent
        )
    except pyone.OneException as e:
        module.fail_json(msg=str(e))

    vm = get_vm_by_id(client, vm_id)
    return get_vm_info(client, vm)


def generate_next_index(vm_filled_indexes_list, num_sign_cnt):
    counter = 0
    cnt_str = str(counter).zfill(num_sign_cnt)

    while cnt_str in vm_filled_indexes_list:
        counter = counter + 1
        cnt_str = str(counter).zfill(num_sign_cnt)

    return cnt_str


def get_vm_labels_and_attributes_dict(client, vm_id):
    vm_USER_TEMPLATE = client.vm.info(vm_id).USER_TEMPLATE

    attrs_dict = {}
    labels_list = []

    for key, value in vm_USER_TEMPLATE.items():
        if key != "LABELS":
            attrs_dict[key] = value
        else:
            if key is not None and value is not None:
                labels_list = value.split(",")

    return labels_list, attrs_dict


def get_all_vms_by_attributes(client, attributes_dict, labels_list):
    pool = client.vmpool.info(-2, -1, -1, -1).VM
    vm_list = []
    name = ""
    if attributes_dict:
        name = attributes_dict.pop("NAME", "")

    if name != "":
        base_name = name[: len(name) - name.count("#")]
        # Check does the name have indexed format
        with_hash = name.endswith("#")

        for vm in pool:
            if vm.NAME.startswith(base_name):
                if with_hash and vm.NAME[len(base_name) :].isdigit():
                    # If the name has indexed format and after base_name it has only digits it'll be matched
                    vm_list.append(vm)
                elif not with_hash and name == vm.NAME:
                    # If the name is not indexed it has to be same
                    vm_list.append(vm)
        pool = vm_list

    vm_list = copy.copy(pool)

    for vm in pool:
        remove_list = []
        vm_labels_list, vm_attributes_dict = get_vm_labels_and_attributes_dict(client, vm.ID)

        if attributes_dict and len(attributes_dict) > 0:
            for key, val in attributes_dict.items():
                if key in vm_attributes_dict:
                    if val and vm_attributes_dict[key] != val:
                        remove_list.append(vm)
                        break
                else:
                    remove_list.append(vm)
                    break
        vm_list = list(set(vm_list).difference(set(remove_list)))

        remove_list = []
        if labels_list and len(labels_list) > 0:
            for label in labels_list:
                if label not in vm_labels_list:
                    remove_list.append(vm)
                    break
        vm_list = list(set(vm_list).difference(set(remove_list)))

    return vm_list


def create_count_of_vms(
    module,
    client,
    template_id,
    count,
    attributes_dict,
    labels_list,
    disk_size,
    network_attrs_list,
    wait,
    wait_timeout,
    vm_start_on_hold,
    vm_persistent,
    updateconf_dict,
):
    new_vms_list = []

    vm_name = ""
    if attributes_dict:
        vm_name = attributes_dict.get("NAME", "")

    if module.check_mode:
        return True, [], []

    # Create list of used indexes
    vm_filled_indexes_list = None
    num_sign_cnt = vm_name.count("#")
    if vm_name != "" and num_sign_cnt > 0:
        vm_list = get_all_vms_by_attributes(client, {"NAME": vm_name}, None)
        base_name = vm_name[: len(vm_name) - num_sign_cnt]
        vm_name = base_name
        # Make list which contains used indexes in format ['000', '001',...]
        vm_filled_indexes_list = [vm.NAME[len(base_name) :].zfill(num_sign_cnt) for vm in vm_list]

    while count > 0:
        new_vm_name = vm_name
        # Create indexed name
        if vm_filled_indexes_list is not None:
            next_index = generate_next_index(vm_filled_indexes_list, num_sign_cnt)
            vm_filled_indexes_list.append(next_index)
            new_vm_name += next_index
        # Update NAME value in the attributes in case there is index
        attributes_dict["NAME"] = new_vm_name
        new_vm_dict = create_vm(
            module,
            client,
            template_id,
            attributes_dict,
            labels_list,
            disk_size,
            network_attrs_list,
            vm_start_on_hold,
            vm_persistent,
            updateconf_dict,
        )
        new_vm_id = new_vm_dict.get("vm_id")
        new_vm = get_vm_by_id(client, new_vm_id)
        new_vms_list.append(new_vm)
        count -= 1

    if vm_start_on_hold:
        if wait:
            for vm in new_vms_list:
                wait_for_hold(module, client, vm, wait_timeout)
    else:
        if wait:
            for vm in new_vms_list:
                wait_for_running(module, client, vm, wait_timeout)

    return True, new_vms_list, []


def create_exact_count_of_vms(
    module,
    client,
    template_id,
    exact_count,
    attributes_dict,
    count_attributes_dict,
    labels_list,
    count_labels_list,
    disk_size,
    network_attrs_list,
    hard,
    wait,
    wait_timeout,
    vm_start_on_hold,
    vm_persistent,
    updateconf_dict,
):
    vm_list = get_all_vms_by_attributes(client, count_attributes_dict, count_labels_list)

    vm_count_diff = exact_count - len(vm_list)
    changed = vm_count_diff != 0

    instances_list = []
    tagged_instances_list = vm_list

    if module.check_mode:
        return changed, instances_list, tagged_instances_list

    if vm_count_diff > 0:
        # Add more VMs
        changed, instances_list, tagged_instances = create_count_of_vms(
            module,
            client,
            template_id,
            vm_count_diff,
            attributes_dict,
            labels_list,
            disk_size,
            network_attrs_list,
            wait,
            wait_timeout,
            vm_start_on_hold,
            vm_persistent,
            updateconf_dict,
        )

        tagged_instances_list += instances_list
    elif vm_count_diff < 0:
        # Delete surplus VMs
        old_vms_list = []

        while vm_count_diff < 0:
            old_vm = vm_list.pop(0)
            old_vms_list.append(old_vm)
            terminate_vm(module, client, old_vm, hard)
            vm_count_diff += 1

        if wait:
            for vm in old_vms_list:
                wait_for_done(module, client, vm, wait_timeout)

        instances_list = old_vms_list
        # store only the remaining instances
        old_vms_set = set(old_vms_list)
        tagged_instances_list = [vm for vm in vm_list if vm not in old_vms_set]

    return changed, instances_list, tagged_instances_list


VM_STATES = [
    "INIT",
    "PENDING",
    "HOLD",
    "ACTIVE",
    "STOPPED",
    "SUSPENDED",
    "DONE",
    "",
    "POWEROFF",
    "UNDEPLOYED",
    "CLONING",
    "CLONING_FAILURE",
]
LCM_STATES = [
    "LCM_INIT",
    "PROLOG",
    "BOOT",
    "RUNNING",
    "MIGRATE",
    "SAVE_STOP",
    "SAVE_SUSPEND",
    "SAVE_MIGRATE",
    "PROLOG_MIGRATE",
    "PROLOG_RESUME",
    "EPILOG_STOP",
    "EPILOG",
    "SHUTDOWN",
    "STATE13",
    "STATE14",
    "CLEANUP_RESUBMIT",
    "UNKNOWN",
    "HOTPLUG",
    "SHUTDOWN_POWEROFF",
    "BOOT_UNKNOWN",
    "BOOT_POWEROFF",
    "BOOT_SUSPENDED",
    "BOOT_STOPPED",
    "CLEANUP_DELETE",
    "HOTPLUG_SNAPSHOT",
    "HOTPLUG_NIC",
    "HOTPLUG_SAVEAS",
    "HOTPLUG_SAVEAS_POWEROFF",
    "HOTPULG_SAVEAS_SUSPENDED",
    "SHUTDOWN_UNDEPLOY",
]


def wait_for_state(module, client, vm, wait_timeout, state_predicate):
    start_time = time.time()

    while (time.time() - start_time) < wait_timeout:
        vm = client.vm.info(vm.ID)
        state = vm.STATE
        lcm_state = vm.LCM_STATE

        if state_predicate(state, lcm_state):
            return vm
        elif state not in [
            VM_STATES.index("INIT"),
            VM_STATES.index("PENDING"),
            VM_STATES.index("HOLD"),
            VM_STATES.index("ACTIVE"),
            VM_STATES.index("CLONING"),
            VM_STATES.index("POWEROFF"),
        ]:
            module.fail_json(msg=f"Action is unsuccessful. VM state: {VM_STATES[state]}")

        time.sleep(1)

    module.fail_json(msg="Wait timeout has expired!")


def wait_for_running(module, client, vm, wait_timeout):
    return wait_for_state(
        module,
        client,
        vm,
        wait_timeout,
        lambda state, lcm_state: state in [VM_STATES.index("ACTIVE")] and lcm_state in [LCM_STATES.index("RUNNING")],
    )


def wait_for_done(module, client, vm, wait_timeout):
    return wait_for_state(module, client, vm, wait_timeout, lambda state, lcm_state: state in [VM_STATES.index("DONE")])


def wait_for_hold(module, client, vm, wait_timeout):
    return wait_for_state(module, client, vm, wait_timeout, lambda state, lcm_state: state in [VM_STATES.index("HOLD")])


def wait_for_poweroff(module, client, vm, wait_timeout):
    return wait_for_state(
        module, client, vm, wait_timeout, lambda state, lcm_state: state in [VM_STATES.index("POWEROFF")]
    )


def terminate_vm(module, client, vm, hard=False):
    changed = False

    if not vm:
        return changed

    changed = True

    if not module.check_mode:
        if hard:
            client.vm.action("terminate-hard", vm.ID)
        else:
            client.vm.action("terminate", vm.ID)

    return changed


def terminate_vms(module, client, vms, hard):
    changed = False

    for vm in vms:
        changed = terminate_vm(module, client, vm, hard) or changed

    return changed


def poweroff_vm(module, client, vm, hard):
    vm = client.vm.info(vm.ID)
    changed = False

    lcm_state = vm.LCM_STATE
    state = vm.STATE

    if lcm_state not in [LCM_STATES.index("SHUTDOWN"), LCM_STATES.index("SHUTDOWN_POWEROFF")] and state not in [
        VM_STATES.index("POWEROFF")
    ]:
        changed = True

    if changed and not module.check_mode:
        if not hard:
            client.vm.action("poweroff", vm.ID)
        else:
            client.vm.action("poweroff-hard", vm.ID)

    return changed


def poweroff_vms(module, client, vms, hard):
    changed = False

    for vm in vms:
        changed = poweroff_vm(module, client, vm, hard) or changed

    return changed


def reboot_vms(module, client, vms, wait_timeout, hard):
    if not module.check_mode:
        # Firstly, power-off all instances
        for vm in vms:
            vm = client.vm.info(vm.ID)
            lcm_state = vm.LCM_STATE
            state = vm.STATE
            if lcm_state not in [LCM_STATES.index("SHUTDOWN_POWEROFF")] and state not in [VM_STATES.index("POWEROFF")]:
                poweroff_vm(module, client, vm, hard)

        # Wait for all to be power-off
        for vm in vms:
            wait_for_poweroff(module, client, vm, wait_timeout)

        for vm in vms:
            resume_vm(module, client, vm)

    return True


def resume_vm(module, client, vm):
    vm = client.vm.info(vm.ID)
    changed = False

    state = vm.STATE
    if state in [VM_STATES.index("HOLD")]:
        changed = release_vm(module, client, vm)
        return changed

    lcm_state = vm.LCM_STATE
    if lcm_state == LCM_STATES.index("SHUTDOWN_POWEROFF"):
        module.fail_json(
            msg="Cannot perform action 'resume' because this action is not available "
            "for LCM_STATE: 'SHUTDOWN_POWEROFF'. Wait for the VM to shutdown properly"
        )
    if lcm_state not in [LCM_STATES.index("RUNNING")]:
        changed = True

    if changed and not module.check_mode:
        client.vm.action("resume", vm.ID)

    return changed


def resume_vms(module, client, vms):
    changed = False

    for vm in vms:
        changed = resume_vm(module, client, vm) or changed

    return changed


def release_vm(module, client, vm):
    vm = client.vm.info(vm.ID)
    changed = False

    state = vm.STATE
    if state != VM_STATES.index("HOLD"):
        module.fail_json(
            msg="Cannot perform action 'release' because this action is not available "
            "because VM is not in state 'HOLD'."
        )
    else:
        changed = True

    if changed and not module.check_mode:
        client.vm.action("release", vm.ID)

    return changed


def check_name_attribute(module, attributes):
    if attributes.get("NAME"):
        if re.match(r"^[^#]+#*$", attributes.get("NAME")) is None:
            module.fail_json(
                msg=f"Illegal 'NAME' attribute: '{attributes.get('NAME')}"
                "' .Signs '#' are allowed only at the end of the name and the name cannot contain only '#'."
            )


TEMPLATE_RESTRICTED_ATTRIBUTES = [
    "CPU",
    "VCPU",
    "OS",
    "FEATURES",
    "MEMORY",
    "DISK",
    "NIC",
    "INPUT",
    "GRAPHICS",
    "CONTEXT",
    "CREATED_BY",
    "CPU_COST",
    "DISK_COST",
    "MEMORY_COST",
    "TEMPLATE_ID",
    "VMID",
    "AUTOMATIC_DS_REQUIREMENTS",
    "DEPLOY_FOLDER",
    "LABELS",
]


def check_attributes(module, attributes):
    for key in attributes.keys():
        if key in TEMPLATE_RESTRICTED_ATTRIBUTES:
            module.fail_json(msg=f"Restricted attribute `{key}` cannot be used when filtering VMs.")
    # Check the format of the name attribute
    check_name_attribute(module, attributes)


def disk_save_as(module, client, vm, disk_saveas, wait_timeout):
    if not disk_saveas.get("name"):
        module.fail_json(msg="Key 'name' is required for 'disk_saveas' option")

    image_name = disk_saveas.get("name")
    disk_id = disk_saveas.get("disk_id", 0)

    if not module.check_mode:
        if VM_STATES.index("POWEROFF") != vm.STATE:
            module.fail_json(msg="'disksaveas' option can be used only when the VM is in 'POWEROFF' state")
        try:
            client.vm.disksaveas(vm.ID, disk_id, image_name, "OS", -1)
        except pyone.OneException as e:
            module.fail_json(msg=str(e))
        wait_for_poweroff(module, client, vm, wait_timeout)  # wait for VM to leave the hotplug_saveas_poweroff state


def get_connection_info(module):
    url = module.params.get("api_url")
    username = module.params.get("api_username")
    password = module.params.get("api_password")

    if not url:
        url = os.environ.get("ONE_URL")

    if not username:
        username = os.environ.get("ONE_USERNAME")

    if not password:
        password = os.environ.get("ONE_PASSWORD")

    if not username:
        if not password:
            authfile = os.environ.get("ONE_AUTH")
            if authfile is None:
                authfile = os.path.join(os.environ.get("HOME"), ".one", "one_auth")
            try:
                with open(authfile) as fp:
                    authstring = fp.read().rstrip()
                username = authstring.split(":")[0]
                password = authstring.split(":")[1]
            except OSError:
                module.fail_json(msg=f"Could not find or read ONE_AUTH file at '{authfile}'")
            except Exception:
                module.fail_json(msg=f"Error occurs when read ONE_AUTH file at '{authfile}'")
    if not url:
        module.fail_json(msg="Opennebula API url (api_url) is not specified")

    return AuthParams(url=url, username=username, password=password)


def main():
    fields = {
        "api_url": {"required": False, "type": "str"},
        "api_username": {"required": False, "type": "str"},
        "api_password": {"required": False, "type": "str", "no_log": True},
        "instance_ids": {"required": False, "aliases": ["ids"], "type": "list", "elements": "int"},
        "template_name": {"required": False, "type": "str"},
        "template_id": {"required": False, "type": "int"},
        "vm_start_on_hold": {"default": False, "type": "bool"},
        "state": {
            "default": "present",
            "choices": ["present", "absent", "rebooted", "poweredoff", "running"],
            "type": "str",
        },
        "mode": {"required": False, "type": "str"},
        "owner_id": {"required": False, "type": "int"},
        "group_id": {"required": False, "type": "int"},
        "wait": {"default": True, "type": "bool"},
        "wait_timeout": {"default": 300, "type": "int"},
        "hard": {"default": False, "type": "bool"},
        "memory": {"required": False, "type": "str"},
        "cpu": {"required": False, "type": "float"},
        "vcpu": {"required": False, "type": "int"},
        "disk_size": {"required": False, "type": "list", "elements": "str"},
        "datastore_name": {"required": False, "type": "str"},
        "datastore_id": {"required": False, "type": "int"},
        "networks": {"default": [], "type": "list", "elements": "dict"},
        "count": {"default": 1, "type": "int"},
        "exact_count": {"required": False, "type": "int"},
        "attributes": {"default": {}, "type": "dict"},
        "count_attributes": {"required": False, "type": "dict"},
        "labels": {"default": [], "type": "list", "elements": "str"},
        "count_labels": {"required": False, "type": "list", "elements": "str"},
        "disk_saveas": {"type": "dict"},
        "persistent": {"default": False, "type": "bool"},
        "updateconf": {"type": "dict"},
    }

    module = AnsibleModule(
        argument_spec=fields,
        mutually_exclusive=[
            ["template_id", "template_name", "instance_ids"],
            ["template_id", "template_name", "disk_saveas"],
            ["instance_ids", "count_attributes", "count"],
            ["instance_ids", "count_labels", "count"],
            ["instance_ids", "exact_count"],
            ["instance_ids", "attributes"],
            ["instance_ids", "labels"],
            ["disk_saveas", "attributes"],
            ["disk_saveas", "labels"],
            ["exact_count", "count"],
            ["count", "hard"],
            ["instance_ids", "cpu"],
            ["instance_ids", "vcpu"],
            ["instance_ids", "memory"],
            ["instance_ids", "disk_size"],
            ["instance_ids", "networks"],
            ["persistent", "disk_size"],
        ],
        supports_check_mode=True,
    )

    if not HAS_PYONE:
        module.fail_json(msg="This module requires pyone to work!")

    auth = get_connection_info(module)
    params = module.params
    instance_ids = params.get("instance_ids")
    requested_template_name = params.get("template_name")
    requested_template_id = params.get("template_id")
    put_vm_on_hold = params.get("vm_start_on_hold")
    state = params.get("state")
    permissions = params.get("mode")
    owner_id = params.get("owner_id")
    group_id = params.get("group_id")
    wait = params.get("wait")
    wait_timeout = params.get("wait_timeout")
    hard = params.get("hard")
    memory = params.get("memory")
    cpu = params.get("cpu")
    vcpu = params.get("vcpu")
    disk_size = params.get("disk_size")
    requested_datastore_id = params.get("datastore_id")
    requested_datastore_name = params.get("datastore_name")
    networks = params.get("networks")
    count = params.get("count")
    exact_count = params.get("exact_count")
    attributes = params.get("attributes")
    count_attributes = params.get("count_attributes")
    labels = params.get("labels")
    count_labels = params.get("count_labels")
    disk_saveas = params.get("disk_saveas")
    persistent = params.get("persistent")
    updateconf = params.get("updateconf")

    if not (auth.username and auth.password):
        module.warn("Credentials missing")
    else:
        one_client = pyone.OneServer(auth.url, session=f"{auth.username}:{auth.password}")

    if attributes:
        attributes = {key.upper(): value for key, value in attributes.items()}
        check_attributes(module, attributes)

    if count_attributes:
        count_attributes = {key.upper(): value for key, value in count_attributes.items()}
        if not attributes:
            module.warn(
                "When you pass `count_attributes` without `attributes` option when deploying, `attributes` option will have same values implicitly."
            )
            attributes = copy.copy(count_attributes)
        check_attributes(module, count_attributes)

    if updateconf:
        check_updateconf(module, updateconf)

    if count_labels and not labels:
        module.warn(
            "When you pass `count_labels` without `labels` option when deploying, `labels` option will have same values implicitly."
        )
        labels = count_labels

    # Fetch template
    template_id = None
    if requested_template_id is not None or requested_template_name:
        template_id = get_template_id(module, one_client, requested_template_id, requested_template_name)
        if template_id is None:
            if requested_template_id is not None:
                module.fail_json(msg=f"There is no template with template_id: {requested_template_id}")
            elif requested_template_name:
                module.fail_json(msg=f"There is no template with name: {requested_template_name}")

    # Fetch datastore
    datastore_id = None
    if requested_datastore_id or requested_datastore_name:
        datastore_id = get_datastore_id(module, one_client, requested_datastore_id, requested_datastore_name)
        if datastore_id is None:
            if requested_datastore_id:
                module.fail_json(msg=f"There is no datastore with datastore_id: {requested_datastore_id}")
            elif requested_datastore_name:
                module.fail_json(msg=f"There is no datastore with name: {requested_datastore_name}")
        else:
            attributes["SCHED_DS_REQUIREMENTS"] = f"ID={datastore_id}"

    if exact_count and template_id is None:
        module.fail_json(msg="Option `exact_count` needs template_id or template_name")

    if exact_count is not None and not (count_attributes or count_labels):
        module.fail_json(
            msg="Either `count_attributes` or `count_labels` has to be specified with option `exact_count`."
        )
    if (count_attributes or count_labels) and exact_count is None:
        module.fail_json(
            msg="Option `exact_count` has to be specified when either `count_attributes` or `count_labels` is used."
        )
    if template_id is not None and state != "present":
        module.fail_json(msg="Only state 'present' is valid for the template")

    if memory:
        attributes["MEMORY"] = str(int(get_size_in_MB(module, memory)))
    if cpu:
        attributes["CPU"] = str(cpu)
    if vcpu:
        attributes["VCPU"] = str(vcpu)

    if exact_count is not None and state != "present":
        module.fail_json(msg="The `exact_count` option is valid only for the `present` state")
    if exact_count is not None and exact_count < 0:
        module.fail_json(msg="`exact_count` cannot be less than 0")
    if count <= 0:
        module.fail_json(msg="`count` has to be greater than 0")

    if permissions is not None:
        if re.match("^[0-7]{3}$", permissions) is None:
            module.fail_json(msg="Option `mode` has to have exactly 3 digits and be in the octet format e.g. 600")

    if exact_count is not None:
        # Deploy an exact count of VMs
        changed, instances_list, tagged_instances_list = create_exact_count_of_vms(
            module,
            one_client,
            template_id,
            exact_count,
            attributes,
            count_attributes,
            labels,
            count_labels,
            disk_size,
            networks,
            hard,
            wait,
            wait_timeout,
            put_vm_on_hold,
            persistent,
            updateconf,
        )
        vms = tagged_instances_list
    elif template_id is not None and state == "present":
        # Deploy count VMs
        changed, instances_list, tagged_instances_list = create_count_of_vms(
            module,
            one_client,
            template_id,
            count,
            attributes,
            labels,
            disk_size,
            networks,
            wait,
            wait_timeout,
            put_vm_on_hold,
            persistent,
            updateconf,
        )
        # instances_list - new instances
        # tagged_instances_list - all instances with specified `count_attributes` and `count_labels`
        vms = instances_list
    else:
        # Fetch data of instances, or change their state
        if not (instance_ids or attributes or labels):
            module.fail_json(msg="At least one of `instance_ids`,`attributes`,`labels` must be passed!")

        if memory or cpu or vcpu or disk_size or networks:
            module.fail_json(
                msg="Parameters as `memory`, `cpu`, `vcpu`, `disk_size` and `networks` you can only set when deploying a VM!"
            )

        if hard and state not in ["rebooted", "poweredoff", "absent", "present"]:
            module.fail_json(
                msg="The 'hard' option can be used only for one of these states: 'rebooted', 'poweredoff', 'absent' and 'present'"
            )

        vms = []
        tagged = False
        changed = False

        if instance_ids:
            vms = get_vms_by_ids(module, one_client, state, instance_ids)
        else:
            tagged = True
            vms = get_all_vms_by_attributes(one_client, attributes, labels)

        if len(vms) == 0 and state != "absent" and state != "present":
            module.fail_json(msg="There are no instances with specified `instance_ids`, `attributes` and/or `labels`")

        if len(vms) == 0 and state == "present" and not tagged:
            module.fail_json(msg="There are no instances with specified `instance_ids`.")

        if tagged and state == "absent":
            module.fail_json(msg="Option `instance_ids` is required when state is `absent`.")

        if state == "absent":
            changed = terminate_vms(module, one_client, vms, hard)
        elif state == "rebooted":
            changed = reboot_vms(module, one_client, vms, wait_timeout, hard)
        elif state == "poweredoff":
            changed = poweroff_vms(module, one_client, vms, hard)
        elif state == "running":
            changed = resume_vms(module, one_client, vms)

        instances_list = vms
        tagged_instances_list = []

    if permissions is not None:
        changed = set_vm_permissions(module, one_client, vms, permissions) or changed

    if owner_id is not None or group_id is not None:
        changed = set_vm_ownership(module, one_client, vms, owner_id, group_id) or changed

    if template_id is None and updateconf is not None:
        changed = update_vms(module, one_client, vms, updateconf) or changed

    if wait and not module.check_mode and state != "present":
        wait_for = {
            "absent": wait_for_done,
            "rebooted": wait_for_running,
            "poweredoff": wait_for_poweroff,
            "running": wait_for_running,
        }
        for vm in vms:
            if vm is not None:
                wait_for[state](module, one_client, vm, wait_timeout)

    if disk_saveas is not None:
        if len(vms) == 0:
            module.fail_json(msg="There is no VM whose disk will be saved.")
        disk_save_as(module, one_client, vms[0], disk_saveas, wait_timeout)
        changed = True

    # instances - a list of instances info whose state is changed or which are fetched with C(instance_ids) option
    instances = list(get_vm_info(one_client, vm) for vm in instances_list if vm is not None)
    instances_ids = list(vm.ID for vm in instances_list if vm is not None)
    # tagged_instances - A list of instances info based on a specific attributes and/or labels that are specified with C(count_attributes) and C(count_labels)
    tagged_instances = list(get_vm_info(one_client, vm) for vm in tagged_instances_list if vm is not None)

    result = {
        "changed": changed,
        "instances": instances,
        "instances_ids": instances_ids,
        "tagged_instances": tagged_instances,
    }

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2016-2017 Hewlett Packard Enterprise Development LP
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: oneview_san_manager
short_description: Manage OneView SAN Manager resources
description:
  - Provides an interface to manage SAN Manager resources. Can create, update, or delete.
requirements:
  - hpOneView >= 3.1.1
author:
  - Felipe Bulsoni (@fgbulsoni)
  - Thiago Miotto (@tmiotto)
  - Adriane Cardozo (@adriane-cardozo)
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    description:
      - Indicates the desired state for the Uplink Set resource.
      - V(present) ensures data properties are compliant with OneView.
      - V(absent) removes the resource from OneView, if it exists.
      - V(connection_information_set) updates the connection information for the SAN Manager. This operation is non-idempotent.
    type: str
    default: present
    choices: [present, absent, connection_information_set]
  data:
    description:
      - List with SAN Manager properties.
    type: dict
    required: true

extends_documentation_fragment:
  - community.general._oneview
  - community.general._oneview.validateetag
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Creates a Device Manager for the Brocade SAN provider with the given hostname and credentials
  community.general.oneview_san_manager:
    config: /etc/oneview/oneview_config.json
    state: present
    data:
      providerDisplayName: Brocade Network Advisor
      connectionInfo:
        - name: Host
          value: 172.18.15.1
        - name: Port
          value: 5989
        - name: Username
          value: username
        - name: Password
          value: password
        - name: UseSsl
          value: true
  delegate_to: localhost

- name: Ensure a Device Manager for the Cisco SAN Provider is present
  community.general.oneview_san_manager:
    config: /etc/oneview/oneview_config.json
    state: present
    data:
      name: 172.18.20.1
      providerDisplayName: Cisco
      connectionInfo:
        - name: Host
          value: 172.18.20.1
        - name: SnmpPort
          value: 161
        - name: SnmpUserName
          value: admin
        - name: SnmpAuthLevel
          value: authnopriv
        - name: SnmpAuthProtocol
          value: sha
        - name: SnmpAuthString
          value: password
  delegate_to: localhost

- name: Sets the SAN Manager connection information
  community.general.oneview_san_manager:
    config: /etc/oneview/oneview_config.json
    state: connection_information_set
    data:
      connectionInfo:
        - name: Host
          value: '172.18.15.1'
        - name: Port
          value: '5989'
        - name: Username
          value: 'username'
        - name: Password
          value: 'password'
        - name: UseSsl
          value: true
  delegate_to: localhost

- name: Refreshes the SAN Manager
  community.general.oneview_san_manager:
    config: /etc/oneview/oneview_config.json
    state: present
    data:
      name: 172.18.15.1
      refreshState: RefreshPending
  delegate_to: localhost

- name: Delete the SAN Manager recently created
  community.general.oneview_san_manager:
    config: /etc/oneview/oneview_config.json
    state: absent
    data:
      name: '172.18.15.1'
  delegate_to: localhost
"""

RETURN = r"""
san_manager:
  description: Has the OneView facts about the SAN Manager.
  returned: On O(state=present). Can be null.
  type: dict
"""

from ansible_collections.community.general.plugins.module_utils._oneview import (
    OneViewModuleBase,
    OneViewModuleValueError,
)


class SanManagerModule(OneViewModuleBase):
    MSG_CREATED = "SAN Manager created successfully."
    MSG_UPDATED = "SAN Manager updated successfully."
    MSG_DELETED = "SAN Manager deleted successfully."
    MSG_ALREADY_PRESENT = "SAN Manager is already present."
    MSG_ALREADY_ABSENT = "SAN Manager is already absent."

    argument_spec = dict(
        state=dict(type="str", default="present", choices=["absent", "present", "connection_information_set"]),
        data=dict(type="dict", required=True),
    )

    def __init__(self):
        super().__init__(additional_arg_spec=self.argument_spec, validate_etag_support=True)
        self.resource_client = self.oneview_client.san_managers

    def execute_module(self):
        if self.data.get("connectionInfo"):
            for connection_hash in self.data.get("connectionInfo"):
                if connection_hash.get("name") == "Host":
                    resource_name = connection_hash.get("value")
        elif self.data.get("name"):
            resource_name = self.data.get("name")
        else:
            msg = 'A "name" or "connectionInfo" must be provided inside the "data" field for this operation. '
            msg += 'If a "connectionInfo" is provided, the "Host" name is considered as the "name" for the resource.'
            raise OneViewModuleValueError(msg)

        resource = self.resource_client.get_by_name(resource_name)

        if self.state == "present":
            changed, msg, san_manager = self._present(resource)
            return dict(changed=changed, msg=msg, ansible_facts=dict(san_manager=san_manager))

        elif self.state == "absent":
            return self.resource_absent(resource, method="remove")

        elif self.state == "connection_information_set":
            changed, msg, san_manager = self._connection_information_set(resource)
            return dict(changed=changed, msg=msg, ansible_facts=dict(san_manager=san_manager))

    def _present(self, resource):
        if not resource:
            provider_uri = self.data.get("providerUri", self._get_provider_uri_by_display_name(self.data))
            return True, self.MSG_CREATED, self.resource_client.add(self.data, provider_uri)
        else:
            merged_data = resource.copy()
            merged_data.update(self.data)

            # Remove 'connectionInfo' from comparison, since it is not possible to validate it.
            resource.pop("connectionInfo", None)
            merged_data.pop("connectionInfo", None)

            if self.compare(resource, merged_data):
                return False, self.MSG_ALREADY_PRESENT, resource
            else:
                updated_san_manager = self.resource_client.update(resource=merged_data, id_or_uri=resource["uri"])
                return True, self.MSG_UPDATED, updated_san_manager

    def _connection_information_set(self, resource):
        if not resource:
            return self._present(resource)
        else:
            merged_data = resource.copy()
            merged_data.update(self.data)
            merged_data.pop("refreshState", None)
            if not self.data.get("connectionInfo", None):
                raise OneViewModuleValueError("A connectionInfo field is required for this operation.")
            updated_san_manager = self.resource_client.update(resource=merged_data, id_or_uri=resource["uri"])
            return True, self.MSG_UPDATED, updated_san_manager

    def _get_provider_uri_by_display_name(self, data):
        display_name = data.get("providerDisplayName")
        provider_uri = self.resource_client.get_provider_uri(display_name)

        if not provider_uri:
            raise OneViewModuleValueError(f"The provider '{display_name}' was not found.")

        return provider_uri


def main():
    SanManagerModule().run()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Ryan Scott Brown <ryansb@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: serverless
short_description: Manages a Serverless Framework project
description:
  - Provides support for managing Serverless Framework (U(https://serverless.com/)) project deployments and stacks.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  state:
    description:
      - Goal state of given stage/project.
    type: str
    choices: [absent, present]
    default: present
  serverless_bin_path:
    description:
      - The path of a serverless framework binary relative to the O(service_path), for example V(node_module/.bin/serverless).
    type: path
  service_path:
    description:
      - The path to the root of the Serverless Service to be operated on.
    type: path
    required: true
  stage:
    description:
      - The name of the serverless framework project stage to deploy to.
      - This uses the serverless framework default "dev".
    type: str
    default: ''
  region:
    description:
      - AWS region to deploy the service to.
      - This parameter defaults to V(us-east-1).
    type: str
    default: ''
  deploy:
    description:
      - Whether or not to deploy artifacts after building them.
      - When this option is V(false) all the functions are built, but no stack update is run to send them out.
      - This is mostly useful for generating artifacts to be stored/deployed elsewhere.
    type: bool
    default: true
  force:
    description:
      - Whether or not to force full deployment, equivalent to serverless C(--force) option.
    type: bool
    default: false
  verbose:
    description:
      - Shows all stack events during deployment, and display any Stack Output.
    type: bool
    default: false
notes:
  - Currently, the C(serverless) command must be in the path of the node executing the task. In the future this may be a flag.
requirements:
  - serverless
  - PyYAML
author:
  - Ryan Scott Brown (@ryansb)
"""

EXAMPLES = r"""
- name: Basic deploy of a service
  community.general.serverless:
    service_path: '{{ project_dir }}'
    state: present

- name: Deploy a project, then pull its resource list back into Ansible
  community.general.serverless:
    stage: dev
    region: us-east-1
    service_path: '{{ project_dir }}'
  register: sls

# The cloudformation stack is always named the same as the full service, so the
# cloudformation_info module can get a full list of the stack resources, as
# well as stack events and outputs
- cloudformation_info:
    region: us-east-1
    stack_name: '{{ sls.service_name }}'
    stack_resources: true

- name: Deploy a project using a locally installed serverless binary
  community.general.serverless:
    stage: dev
    region: us-east-1
    service_path: '{{ project_dir }}'
    serverless_bin_path: node_modules/.bin/serverless
"""

RETURN = r"""
service_name:
  type: str
  description: The service name specified in the serverless.yml that was just deployed.
  returned: always
  sample: my-fancy-service-dev
state:
  type: str
  description: Whether the stack for the serverless project is present/absent.
  returned: always
command:
  type: str
  description: Full C(serverless) command run by this module, in case you want to re-run the command outside the module.
  returned: always
  sample: serverless deploy --stage production
"""

import os

try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

from ansible.module_utils.basic import AnsibleModule


def read_serverless_config(module):
    path = module.params.get("service_path")
    full_path = os.path.join(path, "serverless.yml")

    try:
        with open(full_path) as sls_config:
            config = yaml.safe_load(sls_config.read())
            return config
    except OSError as e:
        module.fail_json(msg=f"Could not open serverless.yml in {full_path}. err: {e}")


def get_service_name(module, stage):
    config = read_serverless_config(module)
    if config.get("service") is None:
        module.fail_json(msg="Could not read `service` key from serverless.yml file")

    if stage:
        return f"{config['service']}-{stage}"

    return f"{config['service']}-{config.get('stage', 'dev')}"


def main():
    module = AnsibleModule(
        argument_spec=dict(
            service_path=dict(type="path", required=True),
            state=dict(type="str", default="present", choices=["absent", "present"]),
            region=dict(type="str", default=""),
            stage=dict(type="str", default=""),
            deploy=dict(type="bool", default=True),
            serverless_bin_path=dict(type="path"),
            force=dict(type="bool", default=False),
            verbose=dict(type="bool", default=False),
        ),
    )

    if not HAS_YAML:
        module.fail_json(msg="yaml is required for this module")

    service_path = module.params.get("service_path")
    state = module.params.get("state")
    region = module.params.get("region")
    stage = module.params.get("stage")
    deploy = module.params.get("deploy", True)
    force = module.params.get("force", False)
    verbose = module.params.get("verbose", False)
    serverless_bin_path = module.params.get("serverless_bin_path")

    if serverless_bin_path is not None:
        command = f"{serverless_bin_path} "
    else:
        command = f"{module.get_bin_path('serverless')} "

    if state == "present":
        command += "deploy "
    elif state == "absent":
        command += "remove "
    else:
        module.fail_json(msg=f"State must either be 'present' or 'absent'. Received: {state}")

    if state == "present":
        if not deploy:
            command += "--noDeploy "
        elif force:
            command += "--force "

    if region:
        command += f"--region {region} "
    if stage:
        command += f"--stage {stage} "
    if verbose:
        command += "--verbose "

    rc, out, err = module.run_command(command, cwd=service_path)
    if rc != 0:
        if state == "absent" and f"-{stage}' does not exist" in out:
            module.exit_json(
                changed=False, state="absent", command=command, out=out, service_name=get_service_name(module, stage)
            )

        module.fail_json(msg=f"Failure when executing Serverless command. Exited {rc}.\nstdout: {out}\nstderr: {err}")

    # gather some facts about the deployment
    module.exit_json(
        changed=True, state="present", out=out, command=command, service_name=get_service_name(module, stage)
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2017, Jasper Lievisse Adriaanse <j@jasper.la>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: vmadm
short_description: Manage SmartOS virtual machines and zones
description:
  - Manage SmartOS virtual machines through vmadm(1M).
author: Jasper Lievisse Adriaanse (@jasperla)
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  archive_on_delete:
    description:
      - When enabled, the zone dataset is mounted on C(/zones/archive) upon removal.
    type: bool
  autoboot:
    description:
      - Whether or not a VM is booted when the system is rebooted.
    type: bool
  brand:
    choices: [joyent, joyent-minimal, lx, kvm, bhyve]
    default: joyent
    description:
      - Type of virtual machine. The V(bhyve) option was added in community.general 0.2.0.
    type: str
  boot:
    description:
      - Set the boot order for KVM VMs.
    type: str
  cpu_cap:
    description:
      - Sets a limit on the amount of CPU time that can be used by a VM. Use V(0) for no cap.
    type: int
  cpu_shares:
    description:
      - Sets a limit on the number of fair share scheduler (FSS) CPU shares for a VM. This limit is relative to all other
        VMs on the system.
    type: int
  cpu_type:
    choices: [qemu64, host]
    default: qemu64
    description:
      - Control the type of virtual CPU exposed to KVM VMs.
    type: str
  customer_metadata:
    description:
      - Metadata to be set and associated with this VM, this contain customer modifiable keys.
    type: dict
  delegate_dataset:
    description:
      - Whether to delegate a ZFS dataset to an OS VM.
    type: bool
  disk_driver:
    description:
      - Default value for a virtual disk model for KVM guests.
    type: str
  disks:
    description:
      - A list of disks to add, valid properties are documented in vmadm(1M).
    type: list
    elements: dict
  dns_domain:
    description:
      - Domain value for C(/etc/hosts).
    type: str
  docker:
    description:
      - Docker images need this flag enabled along with the O(brand) set to C(lx).
    type: bool
  filesystems:
    description:
      - Mount additional filesystems into an OS VM.
    type: list
    elements: dict
  firewall_enabled:
    description:
      - Enables the firewall, allowing fwadm(1M) rules to be applied.
    type: bool
  flexible_disk_size:
    description:
      - This sets an upper bound for the amount of space that a bhyve instance may use for its disks and snapshots of those
        disks (in MiBs).
    type: int
    version_added: 10.5.0
  force:
    description:
      - Force a particular action (in other words, stop or delete a VM).
    type: bool
  fs_allowed:
    description:
      - Comma separated list of filesystem types this zone is allowed to mount.
    type: str
  hostname:
    description:
      - Zone/VM hostname.
    type: str
  image_uuid:
    description:
      - Image UUID.
    type: str
  indestructible_delegated:
    description:
      - Adds an C(@indestructible) snapshot to delegated datasets.
    type: bool
  indestructible_zoneroot:
    description:
      - Adds an C(@indestructible) snapshot to zoneroot.
    type: bool
  internal_metadata:
    description:
      - Metadata to be set and associated with this VM, this contains operator generated keys.
    type: dict
  internal_metadata_namespace:
    description:
      - List of namespaces to be set as C(internal_metadata-only); these namespaces come from O(internal_metadata) rather
        than O(customer_metadata).
    type: str
  kernel_version:
    description:
      - Kernel version to emulate for LX VMs.
    type: str
  limit_priv:
    description:
      - Set (comma separated) list of privileges the zone is allowed to use.
    type: str
  maintain_resolvers:
    description:
      - Resolvers in C(/etc/resolv.conf) are updated when updating the O(resolvers) property.
    type: bool
  max_locked_memory:
    description:
      - Total amount of memory (in MiBs) on the host that can be locked by this VM.
    type: int
  max_lwps:
    description:
      - Maximum number of lightweight processes this VM is allowed to have running.
    type: int
  max_physical_memory:
    description:
      - Maximum amount of memory (in MiBs) on the host that the VM is allowed to use.
    type: int
  max_swap:
    description:
      - Maximum amount of virtual memory (in MiBs) the VM is allowed to use.
    type: int
  mdata_exec_timeout:
    description:
      - Timeout in seconds (or 0 to disable) for the C(svc:/smartdc/mdata:execute) service that runs user-scripts in the zone.
    type: int
  name:
    aliases: [alias]
    description:
      - Name of the VM. vmadm(1M) uses this as an optional name.
    type: str
  nic_driver:
    description:
      - Default value for a virtual NIC model for KVM guests.
    type: str
  nics:
    description:
      - A list of nics to add, valid properties are documented in vmadm(1M).
    type: list
    elements: dict
  nowait:
    description:
      - Consider the provisioning complete when the VM first starts, rather than when the VM has rebooted.
    type: bool
  owner_uuid:
    description:
      - Define the UUID of the owner of the VM.
    type: str
    version_added: 10.5.0
  qemu_opts:
    description:
      - Additional qemu arguments for KVM guests. This overwrites the default arguments provided by vmadm(1M) and should only
        be used for debugging.
    type: str
  qemu_extra_opts:
    description:
      - Additional qemu cmdline arguments for KVM guests.
    type: str
  quota:
    description:
      - Quota on zone filesystems (in MiBs).
    type: int
  ram:
    description:
      - Amount of virtual RAM for a KVM guest (in MiBs).
    type: int
  resolvers:
    description:
      - List of resolvers to be put into C(/etc/resolv.conf).
    type: list
    elements: str
  routes:
    description:
      - Dictionary that maps destinations to gateways, these are set as static routes in the VM.
    type: dict
  spice_opts:
    description:
      - Addition options for SPICE-enabled KVM VMs.
    type: str
  spice_password:
    description:
      - Password required to connect to SPICE. By default no password is set. Please note this can be read from the Global
        Zone.
    type: str
  state:
    choices: [present, running, absent, deleted, stopped, created, restarted, rebooted]
    default: running
    description:
      - States for the VM to be in. Please note that V(present), V(stopped) and V(restarted) operate on a VM that is currently
        provisioned. V(present) means that the VM is created if it was absent, and that it is in a running state. V(absent)
        shutdowns the zone before removing it. V(stopped) means the zone is created if it does not exist already, before shutting
        it down.
    type: str
  tmpfs:
    description:
      - Amount of memory (in MiBs) that is available in the VM for the C(/tmp) filesystem.
    type: int
  uuid:
    description:
      - UUID of the VM. Can either be a full UUID or V(*) for all VMs.
    type: str
  vcpus:
    description:
      - Number of virtual CPUs for a KVM guest.
    type: int
  vga:
    description:
      - Specify VGA emulation used by KVM VMs.
    type: str
  virtio_txburst:
    description:
      - Number of packets that can be sent in a single flush of the tx queue of virtio NICs.
    type: int
  virtio_txtimer:
    description:
      - Timeout (in nanoseconds) for the TX timer of virtio NICs.
    type: int
  vnc_password:
    description:
      - Password required to connect to VNC. By default no password is set. Please note this can be read from the Global Zone.
    type: str
  vnc_port:
    description:
      - TCP port to listen of the VNC server. Or set V(0) for random, or V(-1) to disable.
    type: int
  zfs_data_compression:
    description:
      - Specifies compression algorithm used for this VMs data dataset. This option only has effect on delegated datasets.
    type: str
  zfs_data_recsize:
    description:
      - Suggested block size (power of 2) for files in the delegated dataset's filesystem.
    type: int
  zfs_filesystem_limit:
    description:
      - Maximum number of filesystems the VM can have.
    type: int
  zfs_io_priority:
    description:
      - IO throttle priority value relative to other VMs.
    type: int
  zfs_root_compression:
    description:
      - Specifies compression algorithm used for this VMs root dataset. This option only has effect on the zoneroot dataset.
    type: str
  zfs_root_recsize:
    description:
      - Suggested block size (power of 2) for files in the zoneroot dataset's filesystem.
    type: int
  zfs_snapshot_limit:
    description:
      - Number of snapshots the VM can have.
    type: int
  zpool:
    description:
      - ZFS pool the VM's zone dataset is created in.
    type: str
"""

EXAMPLES = r"""
- name: Create SmartOS zone
  community.general.vmadm:
    brand: joyent
    state: present
    alias: fw_zone
    image_uuid: 95f265b8-96b2-11e6-9597-972f3af4b6d5
    firewall_enabled: true
    indestructible_zoneroot: true
    nics:
      - nic_tag: admin
        ip: dhcp
        primary: true
    internal_metadata:
      root_pw: 'secret'
    quota: 1

- name: Delete a zone
  community.general.vmadm:
    alias: test_zone
    state: deleted

- name: Stop all zones
  community.general.vmadm:
    uuid: '*'
    state: stopped
"""

RETURN = r"""
uuid:
  description: UUID of the managed VM.
  returned: always
  type: str
  sample: 'b217ab0b-cf57-efd8-cd85-958d0b80be33'
alias:
  description: Alias of the managed VM.
  returned: When addressing a VM by alias.
  type: str
  sample: 'dns-zone'
state:
  description: State of the target, after execution.
  returned: success
  type: str
  sample: 'running'
"""

import json
import os
import re
import tempfile
import traceback

from ansible.module_utils.basic import AnsibleModule

# While vmadm(1M) supports a -E option to return any errors in JSON, the
# generated JSON does not play well with the JSON parsers of Python.
# The returned message contains '\n' as part of the stacktrace,
# which breaks the parsers.


def get_vm_prop(module, uuid, prop):
    # Lookup a property for the given VM.
    # Returns the property, or None if not found.
    cmd = [module.vmadm, "lookup", "-j", "-o", prop, f"uuid={uuid}"]

    (rc, stdout, stderr) = module.run_command(cmd)

    if rc != 0:
        module.fail_json(msg=f"Could not perform lookup of {prop} on {uuid}", rc=rc, stdout=stdout, stderr=stderr)

    try:
        stdout_json = json.loads(stdout)
    except Exception as e:
        module.fail_json(
            msg=f"Invalid JSON returned by vmadm for uuid lookup of {prop}",
            details=f"{e}",
            exception=traceback.format_exc(),
        )

    if stdout_json:
        return stdout_json[0].get(prop)


def get_vm_uuid(module, alias):
    # Lookup the uuid that goes with the given alias.
    # Returns the uuid or '' if not found.
    cmd = [module.vmadm, "lookup", "-j", "-o", "uuid", f"alias={alias}"]

    (rc, stdout, stderr) = module.run_command(cmd)

    if rc != 0:
        module.fail_json(msg=f"Could not retrieve UUID of {alias}", rc=rc, stdout=stdout, stderr=stderr)

    # If no VM was found matching the given alias, we get back an empty array.
    # That is not an error condition as we might be explicitly checking for its
    # absence.
    try:
        stdout_json = json.loads(stdout)
    except Exception as e:
        module.fail_json(
            msg=f"Invalid JSON returned by vmadm for uuid lookup of {alias}",
            details=f"{e}",
            exception=traceback.format_exc(),
        )

    if stdout_json:
        return stdout_json[0].get("uuid")


def get_all_vm_uuids(module):
    # Retrieve the UUIDs for all VMs.
    cmd = [module.vmadm, "lookup", "-j", "-o", "uuid"]

    (rc, stdout, stderr) = module.run_command(cmd)

    if rc != 0:
        module.fail_json(msg="Failed to get VMs list", rc=rc, stdout=stdout, stderr=stderr)

    try:
        stdout_json = json.loads(stdout)
        return [v["uuid"] for v in stdout_json]
    except Exception as e:
        module.fail_json(msg="Could not retrieve VM UUIDs", details=f"{e}", exception=traceback.format_exc())


def new_vm(module, uuid, vm_state):
    payload_file = create_payload(module, uuid)

    (rc, stdout, stderr) = vmadm_create_vm(module, payload_file)

    if rc != 0:
        changed = False
        module.fail_json(msg="Could not create VM", rc=rc, stdout=stdout, stderr=stderr)
    else:
        changed = True
        # 'vmadm create' returns all output to stderr...
        match = re.match("Successfully created VM (.*)", stderr)
        if match:
            vm_uuid = match.groups()[0]
            if not is_valid_uuid(vm_uuid):
                module.fail_json(msg=f"Invalid UUID for VM {vm_uuid}?")
        else:
            module.fail_json(msg="Could not retrieve UUID of newly created(?) VM")

        # Now that the VM is created, ensure it is in the desired state (if not 'running')
        if vm_state != "running":
            ret = set_vm_state(module, vm_uuid, vm_state)
            if not ret:
                module.fail_json(msg=f"Could not set VM {vm_uuid} to state {vm_state}")

    try:
        os.unlink(payload_file)
    except Exception as e:
        # Since the payload may contain sensitive information, fail hard
        # if we cannot remove the file so the operator knows about it.
        module.fail_json(
            msg=f"Could not remove temporary JSON payload file {payload_file}: {e}", exception=traceback.format_exc()
        )

    return changed, vm_uuid


def vmadm_create_vm(module, payload_file):
    # Create a new VM using the provided payload.
    cmd = [module.vmadm, "create", "-f", payload_file]

    return module.run_command(cmd)


def set_vm_state(module, vm_uuid, vm_state):
    p = module.params

    # Check if the VM is already in the desired state.
    state = get_vm_prop(module, vm_uuid, "state")
    if state and (state == vm_state):
        return None

    # Lookup table for the state to be in, and which command to use for that.
    # vm_state: [vmadm commandm, forceable?]
    cmds = {
        "stopped": ["stop", True],
        "running": ["start", False],
        "deleted": ["delete", True],
        "rebooted": ["reboot", False],
    }

    command, forceable = cmds[vm_state]
    force = ["-F"] if p["force"] and forceable else []

    cmd = [module.vmadm, command] + force + [vm_uuid]

    (dummy, dummy, stderr) = module.run_command(cmd)

    match = re.match("^Successfully.*", stderr)
    return match is not None


def create_payload(module, uuid):
    # Create the JSON payload (vmdef) and return the filename.

    # Filter out the few options that are not valid VM properties.
    module_options = ["force", "state"]
    vmdef = {k: v for k, v in module.params.items() if k not in module_options and v}

    try:
        vmdef_json = json.dumps(vmdef)
    except Exception:
        module.fail_json(msg="Could not create valid JSON payload", exception=traceback.format_exc())

    # Create the temporary file that contains our payload, and set tight
    # permissions for it may container sensitive information.
    try:
        # XXX: When there's a way to get the current ansible temporary directory
        # drop the mkstemp call and rely on ANSIBLE_KEEP_REMOTE_FILES to retain
        # the payload (thus removing the `save_payload` option).
        fname = tempfile.mkstemp()[1]
        os.chmod(fname, 0o400)
        with open(fname, "w") as fh:
            fh.write(vmdef_json)
    except Exception as e:
        module.fail_json(msg=f"Could not save JSON payload: {e}", exception=traceback.format_exc())

    return fname


def vm_state_transition(module, uuid, vm_state):
    ret = set_vm_state(module, uuid, vm_state)

    # Whether the VM changed state.
    if ret is None:
        return False
    elif ret:
        return True
    else:
        module.fail_json(msg=f"Failed to set VM {uuid} to state {vm_state}")


def is_valid_uuid(uuid):
    return re.match("^[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$", uuid, re.IGNORECASE) is not None


def validate_uuids(module):
    failed = [
        name
        for name, pvalue in [(x, module.params[x]) for x in ["uuid", "image_uuid"]]
        if pvalue and pvalue != "*" and not is_valid_uuid(pvalue)
    ]

    if failed:
        module.fail_json(msg=f"No valid UUID(s) found for: {', '.join(failed)}")


def manage_all_vms(module, vm_state):
    # Handle operations for all VMs, which can by definition only
    # be state transitions.
    state = module.params["state"]

    if state == "created":
        module.fail_json(msg='State "created" is only valid for tasks with a single VM')

    # If any of the VMs has a change, the task as a whole has a change.
    any_changed = False

    # First get all VM uuids and for each check their state, and adjust it if needed.
    for uuid in get_all_vm_uuids(module):
        current_vm_state = get_vm_prop(module, uuid, "state")
        if not current_vm_state and vm_state == "deleted":
            any_changed = False
        else:
            if module.check_mode:
                if (not current_vm_state) or (get_vm_prop(module, uuid, "state") != state):
                    any_changed = True
            else:
                any_changed = vm_state_transition(module, uuid, vm_state) or any_changed

    return any_changed


def main():
    # In order to reduce the clutter and boilerplate for trivial options,
    # abstract the vmadm properties and build the dict of arguments later.
    # Dict of all options that are simple to define based on their type.
    # They're not required and have a default of None.
    properties = {
        "str": [
            "boot",
            "disk_driver",
            "dns_domain",
            "fs_allowed",
            "hostname",
            "image_uuid",
            "internal_metadata_namespace",
            "kernel_version",
            "limit_priv",
            "nic_driver",
            "owner_uuid",
            "qemu_opts",
            "qemu_extra_opts",
            "spice_opts",
            "uuid",
            "vga",
            "zfs_data_compression",
            "zfs_root_compression",
            "zpool",
        ],
        "bool": [
            "archive_on_delete",
            "autoboot",
            "delegate_dataset",
            "docker",
            "firewall_enabled",
            "force",
            "indestructible_delegated",
            "indestructible_zoneroot",
            "maintain_resolvers",
            "nowait",
        ],
        "int": [
            "cpu_cap",
            "cpu_shares",
            "flexible_disk_size",
            "max_locked_memory",
            "max_lwps",
            "max_physical_memory",
            "max_swap",
            "mdata_exec_timeout",
            "quota",
            "ram",
            "tmpfs",
            "vcpus",
            "virtio_txburst",
            "virtio_txtimer",
            "vnc_port",
            "zfs_data_recsize",
            "zfs_filesystem_limit",
            "zfs_io_priority",
            "zfs_root_recsize",
            "zfs_snapshot_limit",
        ],
        "dict": ["customer_metadata", "internal_metadata", "routes"],
    }

    # Start with the options that are not as trivial as those above.
    options = dict(
        state=dict(
            default="running",
            type="str",
            choices=["present", "running", "absent", "deleted", "stopped", "created", "restarted", "rebooted"],
        ),
        name=dict(type="str", aliases=["alias"]),
        brand=dict(default="joyent", type="str", choices=["joyent", "joyent-minimal", "lx", "kvm", "bhyve"]),
        cpu_type=dict(default="qemu64", type="str", choices=["host", "qemu64"]),
        # Regular strings, however these require additional options.
        spice_password=dict(type="str", no_log=True),
        vnc_password=dict(type="str", no_log=True),
        disks=dict(type="list", elements="dict"),
        nics=dict(type="list", elements="dict"),
        resolvers=dict(type="list", elements="str"),
        filesystems=dict(type="list", elements="dict"),
    )

    # Add our 'simple' options to options dict.
    for type in properties:
        for p in properties[type]:
            option = dict(type=type)
            options[p] = option

    module = AnsibleModule(argument_spec=options, supports_check_mode=True, required_one_of=[["name", "uuid"]])

    module.vmadm = module.get_bin_path("vmadm", required=True)

    p = module.params
    uuid = p["uuid"]
    state = p["state"]

    # Translate the state parameter into something we can use later on.
    if state in ["present", "running"]:
        vm_state = "running"
    elif state in ["stopped", "created"]:
        vm_state = "stopped"
    elif state in ["absent", "deleted"]:
        vm_state = "deleted"
    elif state in ["restarted", "rebooted"]:
        vm_state = "rebooted"

    result = {"state": state}

    # While it is possible to refer to a given VM by its `alias`, it is easier
    # to operate on VMs by their UUID. So if we're not given a `uuid`, look
    # it up.
    if not uuid:
        uuid = get_vm_uuid(module, p["name"])
        # Bit of a chicken and egg problem here for VMs with state == deleted.
        # If they're going to be removed in this play, we have to lookup the
        # uuid. If they're already deleted there's nothing to lookup.
        # So if state == deleted and get_vm_uuid() returned '', the VM is already
        # deleted and there's nothing else to do.
        if uuid is None and vm_state == "deleted":
            result["name"] = p["name"]
            module.exit_json(**result)

    validate_uuids(module)

    if p["name"]:
        result["name"] = p["name"]
    result["uuid"] = uuid

    if uuid == "*":
        result["changed"] = manage_all_vms(module, vm_state)
        module.exit_json(**result)

    # The general flow is as follows:
    # - first the current state of the VM is obtained by its UUID.
    # - If the state was not found and the desired state is 'deleted', return.
    # - If the state was not found, it means the VM has to be created.
    #   Subsequently the VM will be set to the desired state (i.e. stopped)
    # - Otherwise, it means the VM exists already and we operate on its
    #   state (for example reboot it.)
    #
    # In the future it should be possible to query the VM for a particular
    # property as a valid state (i.e. queried) so the result can be
    # registered.
    # Also, VMs should be able to get their properties updated.
    # Managing VM snapshots should be part of a standalone module.

    # First obtain the VM state to determine what needs to be done with it.
    current_vm_state = get_vm_prop(module, uuid, "state")

    # First handle the case where the VM should be deleted and is not present.
    if not current_vm_state and vm_state == "deleted":
        result["changed"] = False
    elif module.check_mode:
        # Shortcut for check mode, if there is no VM yet, it will need to be created.
        # Or, if the VM is not in the desired state yet, it needs to transition.
        result["changed"] = (not current_vm_state) or (get_vm_prop(module, uuid, "state") != state)
    elif not current_vm_state:
        # No VM was found that matched the given ID (alias or uuid), so we create it.
        result["changed"], result["uuid"] = new_vm(module, uuid, vm_state)
    else:
        # VM was found, operate on its state directly.
        result["changed"] = vm_state_transition(module, uuid, vm_state)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Artem Feofanov <artem.feofanov@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: telegram
author:
  - "Artem Feofanov (@tyouxa)"
  - "Nikolai Lomov (@lomserman)"

short_description: Send notifications using telegram

description:
  - Send notifications using telegram bot, to a verified group or user.
  - Also, the user may try to use any other telegram bot API method, if you specify O(api_method) argument.
notes:
  - You need a telegram account and create telegram bot to use this module.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  token:
    type: str
    description:
      - Token identifying your telegram bot.
    required: true
  api_host:
    type: str
    description:
      - Custom telegram API host.
    default: api.telegram.org
    version_added: 13.0.0
  api_method:
    type: str
    description:
      - Bot API method.
      - For reference, see U(https://core.telegram.org/bots/api).
    default: SendMessage
    version_added: 2.0.0
  api_args:
    type: dict
    description:
      - Any parameters for the method.
      - For reference to default method, V(SendMessage), see U(https://core.telegram.org/bots/api#sendmessage).
    version_added: 2.0.0
"""

EXAMPLES = r"""
- name: Send notify to Telegram
  community.general.telegram:
    token: '9999999:XXXXXXXXXXXXXXXXXXXXXXX'
    api_args:
      chat_id: "000000"
      parse_mode: "markdown"
      text: "Your precious application has been deployed: https://example.com"
      disable_web_page_preview: true
      disable_notification: true

- name: Forward message to someone
  community.general.telegram:
    token: '9999999:XXXXXXXXXXXXXXXXXXXXXXX'
    api_method: forwardMessage
    api_args:
      chat_id: "000000"
      from_chat_id: 111111
      disable_notification: true
      message_id: '{{ saved_msg_id }}'

- name: Send notify to custom telegram API host
  community.general.telegram:
    token: '9999999:XXXXXXXXXXXXXXXXXXXXXXX'
    api_host: "telegram.example.com"
    api_args:
      chat_id: "000000"
      parse_mode: "markdown"
      text: "Your precious application has been deployed: https://example.com"
      disable_web_page_preview: true
      disable_notification: true
"""

RETURN = r"""
msg:
  description: The message you attempted to send.
  returned: success
  type: str
  sample: "Ansible task finished"
telegram_error:
  description: Error message gotten from Telegram API.
  returned: failure
  type: str
  sample: "Bad Request: message text is empty"
"""

import json
from urllib.parse import quote

from ansible.module_utils.basic import AnsibleModule

# noinspection PyUnresolvedReferences
from ansible.module_utils.urls import fetch_url


def main():
    module = AnsibleModule(
        argument_spec=dict(
            token=dict(type="str", required=True, no_log=True),
            api_host=dict(type="str", default="api.telegram.org"),
            api_args=dict(type="dict"),
            api_method=dict(type="str", default="SendMessage"),
        ),
        supports_check_mode=True,
    )

    token = quote(module.params["token"])
    api_host = module.params["api_host"]
    api_args = module.params["api_args"] or {}
    api_method = module.params["api_method"]
    # filling backward compatibility args
    api_args["chat_id"] = api_args.get("chat_id")
    api_args["parse_mode"] = api_args.get("parse_mode")
    api_args["text"] = api_args.get("text")

    if api_args["parse_mode"] == "plain":
        del api_args["parse_mode"]

    url = f"https://{api_host}/bot{token}/{api_method}"

    if module.check_mode:
        module.exit_json(changed=False)

    response, info = fetch_url(
        module, url, method="POST", data=json.dumps(api_args), headers={"Content-Type": "application/json"}
    )
    if info["status"] == 200:
        module.exit_json(changed=True)
    elif info["status"] == -1:
        # SSL errors, connection problems, etc.
        module.fail_json(msg="Failed to send message", info=info, response=response)
    else:
        body = json.loads(info["body"])
        module.fail_json(
            msg=f"Failed to send message, return status = {info['status']}\nurl = {url}\napi_args = {api_args}",
            telegram_error=body["description"],
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_summary_top_applications_categories_by_usage_info
short_description: Information module for organizations _summary _top _applications
  _categories _by _usage
description:
  - Information module for Organizations Summary Top Applications Categories By Usage Info.
  - Get all organizations _summary _top _applications _categories _by _usage.
  - Return the top application categories sorted by data usage over given time range.
    Default unit is megabytes.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - OrganizationId path parameter. Organization ID.
    type: str
  networkTag:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - NetworkTag query parameter. Match result to an exact network tag.
    type: str
  deviceTag:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - DeviceTag query parameter. Match result to an exact device tag.
    type: str
  networkId:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - NetworkId query parameter. Match result to an exact network id.
    type: str
  quantity:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - Quantity query parameter. Set number of desired results to return. Default
        is 10. Maximum is 50.
    type: int
  ssidName:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - SsidName query parameter. Filter results by ssid name.
    type: str
  usageUplink:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - UsageUplink query parameter. Filter results by usage uplink.
    type: str
  t0:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - T0 query parameter. The beginning of the timespan for the data.
    type: str
  t1:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - T1 query parameter. The end of the timespan for the data. T1 can be a maximum
        of 186 days after t0.
    type: str
  timespan:
    description:
      - Information module for Organizations Summary Top Applications Categories By Usage Info.
      - >
        Timespan query parameter. The timespan for which the information will be fetched.
        If specifying timespan, do not specify parameters t0 and t1. The value must
        be in seconds and be greater than or equal to 25 minutes and be less than
        or equal to 186 days. The default is 1 day.
    type: float
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations getOrganizationSummaryTopApplicationsCategoriesByUsage
    description: Complete reference of the getOrganizationSummaryTopApplicationsCategoriesByUsage
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-summary-top-applications-categories-by-usage
notes:
  - SDK Method used are
    organizations.Organizations.get_organization_summary_top_applications_categories_by_usage,
  - Paths used are
    get /organizations/{organizationId}/summary/top/applications/categories/byUsage,
"""

EXAMPLES = r"""
- name: Get all organizations _summary _top _applications _categories _by _usage
  cisco.meraki.organizations_summary_top_applications_categories_by_usage_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkTag: string
    deviceTag: string
    networkId: string
    quantity: 0
    ssidName: string
    usageUplink: string
    t0: string
    t1: string
    timespan: 0
    organizationId: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "category": "string",
        "total": 0,
        "downstream": 0,
        "upstream": 0,
        "percentage": 0
      }
    ]
"""



#!/usr/bin/python
# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: online_user_info
short_description: Gather information about Online user
description:
  - Gather information about the user.
author:
  - "Remy Leone (@remyleone)"
extends_documentation_fragment:
  - community.general._online
  - community.general._attributes
  - community.general._attributes.info_module
"""

EXAMPLES = r"""
- name: Gather Online user info
  community.general.online_user_info:
  register: result

- ansible.builtin.debug:
    msg: "{{ result.online_user_info }}"
"""

RETURN = r"""
online_user_info:
  description:
    - Response from Online API.
    - 'For more details please refer to: U(https://console.online.net/en/api/).'
  returned: success
  type: dict
  sample:
    {
      "company": "foobar LLC",
      "email": "foobar@example.com",
      "first_name": "foo",
      "id": 42,
      "last_name": "bar",
      "login": "foobar"
    }
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._online import (
    Online,
    OnlineException,
    online_argument_spec,
)


class OnlineUserInfo(Online):
    def __init__(self, module):
        super().__init__(module)
        self.name = "api/v1/user"


def main():
    module = AnsibleModule(
        argument_spec=online_argument_spec(),
        supports_check_mode=True,
    )

    try:
        module.exit_json(online_user_info=OnlineUserInfo(module).get_resources())
    except OnlineException as exc:
        module.fail_json(msg=exc.message)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2021, Florian Dambrine <android.florian@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: pritunl_org
author: Florian Dambrine (@Lowess)
version_added: 2.5.0
short_description: Manages Pritunl Organizations using the Pritunl API
description:
  - A module to manage Pritunl organizations using the Pritunl API.
extends_documentation_fragment:
  - community.general._pritunl
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    type: str
    required: true
    aliases:
      - org
    description:
      - The name of the organization to manage in Pritunl.
  force:
    type: bool
    default: false
    description:
      - If O(force) is V(true) and O(state) is V(absent), the module deletes the organization, no matter if it contains users
        or not. By default O(force) is V(false), which causes the module to fail the deletion of the organization when it
        contains users.
  state:
    type: str
    default: 'present'
    choices:
      - present
      - absent
    description:
      - If V(present), the module adds organization O(name) to Pritunl. If V(absent), attempt to delete the organization from
        Pritunl (please read about O(force) usage).
"""

EXAMPLES = r"""
- name: Ensure the organization named MyOrg exists
  community.general.pritunl_org:
    state: present
    name: MyOrg

- name: Ensure the organization named MyOrg does not exist
  community.general.pritunl_org:
    state: absent
    name: MyOrg
"""

RETURN = r"""
response:
  description: JSON representation of a Pritunl Organization.
  returned: success
  type: dict
  sample:
    {
      "auth_api": false,
      "name": "Foo",
      "auth_token": null,
      "user_count": 0,
      "auth_secret": null,
      "id": "csftwlu6uhralzi2dpmhekz3"
    }
"""


from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.dict_transformations import dict_merge

from ansible_collections.community.general.plugins.module_utils._pritunl_api import (
    PritunlException,
    delete_pritunl_organization,
    get_pritunl_settings,
    list_pritunl_organizations,
    post_pritunl_organization,
    pritunl_argument_spec,
)


def add_pritunl_organization(module):
    result = {}

    org_name = module.params.get("name")

    org_obj_list = list_pritunl_organizations(
        **dict_merge(
            get_pritunl_settings(module),
            {"filters": {"name": org_name}},
        )
    )

    # If the organization already exists
    if len(org_obj_list) > 0:
        result["changed"] = False
        result["response"] = org_obj_list[0]
    else:
        # Otherwise create it
        response = post_pritunl_organization(
            **dict_merge(
                get_pritunl_settings(module),
                {"organization_name": org_name},
            )
        )
        result["changed"] = True
        result["response"] = response

    module.exit_json(**result)


def remove_pritunl_organization(module):
    result = {}

    org_name = module.params.get("name")
    force = module.params.get("force")

    org_obj_list = []

    org_obj_list = list_pritunl_organizations(
        **dict_merge(
            get_pritunl_settings(module),
            {
                "filters": {"name": org_name},
            },
        )
    )

    # No organization found
    if len(org_obj_list) == 0:
        result["changed"] = False
        result["response"] = {}

    else:
        # Otherwise attempt to delete it
        org = org_obj_list[0]

        # Only accept deletion under specific conditions
        if force or org["user_count"] == 0:
            response = delete_pritunl_organization(
                **dict_merge(
                    get_pritunl_settings(module),
                    {"organization_id": org["id"]},
                )
            )
            result["changed"] = True
            result["response"] = response
        else:
            module.fail_json(
                msg=(
                    f"Can not remove organization '{org_name}' with {org['user_count']} attached users. "
                    "Either set 'force' option to true or remove active users "
                    "from the organization"
                )
            )

    module.exit_json(**result)


def main():
    argument_spec = pritunl_argument_spec()

    argument_spec.update(
        dict(
            name=dict(required=True, type="str", aliases=["org"]),
            force=dict(type="bool", default=False),
            state=dict(choices=["present", "absent"], default="present"),
        )
    )

    module = AnsibleModule(argument_spec=argument_spec)

    state = module.params.get("state")

    try:
        if state == "present":
            add_pritunl_organization(module)
        elif state == "absent":
            remove_pritunl_organization(module)
    except PritunlException as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2016, Tomas Karasek <tom.to.the.k@gmail.com>
# Copyright (c) 2016, Matt Baldwin <baldwin@stackpointcloud.com>
# Copyright (c) 2016, Thibaud Morel l'Horset <teebes@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: packet_device

short_description: Manage a bare metal server in the Packet Host

description:
  - Manage a bare metal server in the Packet Host (a "device" in the API terms).
  - When the machine is created it can optionally wait for public IP address, or for active state.
  - This module has a dependency on packet >= 1.0.
  - API is documented at U(https://www.packet.net/developers/api/devices).
author:
  - Tomas Karasek (@t0mk) <tom.to.the.k@gmail.com>
  - Matt Baldwin (@baldwinSPC) <baldwin@stackpointcloud.com>
  - Thibaud Morel l'Horset (@teebes) <teebes@gmail.com>

extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  auth_token:
    description:
      - Packet API token. You can also supply it in environment variable E(PACKET_API_TOKEN).
    type: str

  count:
    description:
      - The number of devices to create. Count number can be included in hostname using the C(%d) string formatter.
    default: 1
    type: int

  count_offset:
    description:
      - From which number to start the count.
    default: 1
    type: int

  device_ids:
    description:
      - List of device IDs on which to operate.
    type: list
    elements: str

  tags:
    description:
      - List of device tags.
      - Currently implemented only for device creation.
    type: list
    elements: str
    version_added: '0.2.0'

  facility:
    description:
      - Facility slug for device creation. See Packet API for current list - U(https://www.packet.net/developers/api/facilities/).
    type: str

  features:
    description:
      - Dict with "features" for device creation. See Packet API docs for details.
    type: dict

  hostnames:
    description:
      - A hostname of a device, or a list of hostnames.
      - If given string or one-item list, you can use the C("%d") Python string format to expand numbers from O(count).
      - If only one hostname, it might be expanded to list if O(count)>1.
    aliases: [name]
    type: list
    elements: str

  locked:
    description:
      - Whether to lock a created device.
    default: false
    aliases: [lock]
    type: bool

  operating_system:
    description:
      - OS slug for device creation. See Packet API for current list - U(https://www.packet.net/developers/api/operatingsystems/).
    type: str

  plan:
    description:
      - Plan slug for device creation. See Packet API for current list - U(https://www.packet.net/developers/api/plans/).
    type: str

  project_id:
    description:
      - ID of project of the device.
    required: true
    type: str

  state:
    description:
      - Desired state of the device.
      - If set to V(present) (the default), the module call returns immediately after the device-creating HTTP request successfully
        returns.
      - If set to V(active), the module call blocks until all the specified devices are in state active due to the Packet
        API, or until O(wait_timeout).
    choices: [present, absent, active, inactive, rebooted]
    default: present
    type: str

  user_data:
    description:
      - Userdata blob made available to the machine.
    type: str

  wait_for_public_IPv:
    description:
      - Whether to wait for the instance to be assigned a public IPv4/IPv6 address.
      - If set to V(4), it waits until IPv4 is assigned to the instance.
      - If set to V(6), it waits until public IPv6 is assigned to the instance.
    choices: [4, 6]
    type: int

  wait_timeout:
    description:
      - How long (seconds) to wait either for automatic IP address assignment, or for the device to reach the V(active) state.
      - If O(wait_for_public_IPv) is set and O(state=active), the module waits for both events consequently, applying the
        timeout twice.
    default: 900
    type: int

  ipxe_script_url:
    description:
      - URL of custom iPXE script for provisioning.
      - More about custom iPXE for Packet devices at U(https://help.packet.net/technical/infrastructure/custom-ipxe).
    type: str
    default: ''

  always_pxe:
    description:
      - Persist PXE as the first boot option.
      - Normally, the PXE process happens only on the first boot. Set this arg to have your device continuously boot to iPXE.
    default: false
    type: bool


requirements:
  - "packet-python >= 1.35"
"""

EXAMPLES = r"""
# All the examples assume that you have your Packet API token in environment variable PACKET_API_TOKEN.
# You can also pass it to the auth_token parameter of the module instead.

# Creating devices

- name: Create 1 device
  hosts: localhost
  tasks:
    - community.general.packet_device:
        project_id: 89b497ee-5afc-420a-8fb5-56984898f4df
        hostnames: myserver
        tags: ci-xyz
        operating_system: ubuntu_16_04
        plan: baremetal_0
        facility: sjc1

# Create the same device and wait until it is in state "active", (when it is
# ready for other API operations). Fail if the device is not "active" in
# 10 minutes.

- name: Create device and wait up to 10 minutes for active state
  hosts: localhost
  tasks:
    - community.general.packet_device:
        project_id: 89b497ee-5afc-420a-8fb5-56984898f4df
        hostnames: myserver
        operating_system: ubuntu_16_04
        plan: baremetal_0
        facility: sjc1
        state: active
        wait_timeout: 600

- name: Create 3 ubuntu devices called server-01, server-02 and server-03
  hosts: localhost
  tasks:
    - community.general.packet_device:
        project_id: 89b497ee-5afc-420a-8fb5-56984898f4df
        hostnames: server-%02d
        count: 3
        operating_system: ubuntu_16_04
        plan: baremetal_0
        facility: sjc1

- name: Create 3 coreos devices with userdata, wait until they get IPs and then wait for SSH
  hosts: localhost
  tasks:
    - name: Create 3 devices and register their facts
      community.general.packet_device:
        hostnames: [coreos-one, coreos-two, coreos-three]
        operating_system: coreos_stable
        plan: baremetal_0
        facility: ewr1
        locked: true
        project_id: 89b497ee-5afc-420a-8fb5-56984898f4df
        wait_for_public_IPv: 4
        user_data: |
          #cloud-config
          ssh_authorized_keys:
            - {{ lookup('file', 'my_packet_sshkey') }}
          coreos:
            etcd:
              discovery: https://discovery.etcd.io/6a28e078895c5ec737174db2419bb2f3
              addr: $private_ipv4:4001
              peer-addr: $private_ipv4:7001
            fleet:
              public-ip: $private_ipv4
            units:
              - name: etcd.service
                command: start
              - name: fleet.service
                command: start
      register: newhosts

    - name: Wait for ssh
      ansible.builtin.wait_for:
        delay: 1
        host: "{{ item.public_ipv4 }}"
        port: 22
        state: started
        timeout: 500
      with_items: "{{ newhosts.devices }}"


# Other states of devices

- name: Remove 3 devices by uuid
  hosts: localhost
  tasks:
    - community.general.packet_device:
        project_id: 89b497ee-5afc-420a-8fb5-56984898f4df
        state: absent
        device_ids:
          - 1fb4faf8-a638-4ac7-8f47-86fe514c30d8
          - 2eb4faf8-a638-4ac7-8f47-86fe514c3043
          - 6bb4faf8-a638-4ac7-8f47-86fe514c301f
"""

RETURN = r"""
devices:
  description: Information about each device that was processed.
  type: list
  sample:
    - "hostname": "my-server.com"
      "id": "2a5122b9-c323-4d5c-b53c-9ad3f54273e7"
      "public_ipv4": "147.229.15.12"
      "private-ipv4": "10.0.15.12"
      "tags": []
      "locked": false
      "state": "provisioning"
      "public_ipv6": "2604:1380:2:5200::3"
  returned: success
"""


import os
import re
import time
import traceback
import uuid

from ansible.module_utils.basic import AnsibleModule

HAS_PACKET_SDK = True
try:
    import packet
except ImportError:
    HAS_PACKET_SDK = False


NAME_RE = r"([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])"
HOSTNAME_RE = rf"({NAME_RE}\.)*{NAME_RE}$"
MAX_DEVICES = 100

PACKET_DEVICE_STATES = (
    "queued",
    "provisioning",
    "failed",
    "powering_on",
    "active",
    "powering_off",
    "inactive",
    "rebooting",
)

PACKET_API_TOKEN_ENV_VAR = "PACKET_API_TOKEN"


ALLOWED_STATES = ["absent", "active", "inactive", "rebooted", "present"]


def serialize_device(device):
    """
    Standard representation for a device as returned by various tasks::

        {
            'id': 'device_id'
            'hostname': 'device_hostname',
            'tags': [],
            'locked': false,
            'state': 'provisioning',
            'ip_addresses': [
                {
                    "address": "147.75.194.227",
                    "address_family": 4,
                    "public": true
                },
                {
                    "address": "2604:1380:2:5200::3",
                    "address_family": 6,
                    "public": true
                },
                {
                    "address": "10.100.11.129",
                    "address_family": 4,
                    "public": false
                }
            ],
            "private_ipv4": "10.100.11.129",
            "public_ipv4": "147.75.194.227",
            "public_ipv6": "2604:1380:2:5200::3",
        }

    """
    device_data = {}
    device_data["id"] = device.id
    device_data["hostname"] = device.hostname
    device_data["tags"] = device.tags
    device_data["locked"] = device.locked
    device_data["state"] = device.state
    device_data["ip_addresses"] = [
        {
            "address": addr_data["address"],
            "address_family": addr_data["address_family"],
            "public": addr_data["public"],
        }
        for addr_data in device.ip_addresses
    ]
    # Also include each IPs as a key for easier lookup in roles.
    # Key names:
    # - public_ipv4
    # - public_ipv6
    # - private_ipv4
    # - private_ipv6 (if there is one)
    for ipdata in device_data["ip_addresses"]:
        if ipdata["public"]:
            if ipdata["address_family"] == 6:
                device_data["public_ipv6"] = ipdata["address"]
            elif ipdata["address_family"] == 4:
                device_data["public_ipv4"] = ipdata["address"]
        elif not ipdata["public"]:
            if ipdata["address_family"] == 6:
                # Packet doesn't give public ipv6 yet, but maybe one
                # day they will
                device_data["private_ipv6"] = ipdata["address"]
            elif ipdata["address_family"] == 4:
                device_data["private_ipv4"] = ipdata["address"]
    return device_data


def is_valid_hostname(hostname):
    return re.match(HOSTNAME_RE, hostname) is not None


def is_valid_uuid(myuuid):
    try:
        val = uuid.UUID(myuuid, version=4)
    except ValueError:
        return False
    return str(val) == myuuid


def listify_string_name_or_id(s):
    if "," in s:
        return s.split(",")
    else:
        return [s]


def get_hostname_list(module):
    # hostname is a list-typed param, so I guess it should return list
    # (and it does, in Ansible 2.2.1) but in order to be defensive,
    # I keep here the code to convert an eventual string to list
    hostnames = module.params.get("hostnames")
    count = module.params.get("count")
    count_offset = module.params.get("count_offset")
    if isinstance(hostnames, str):
        hostnames = listify_string_name_or_id(hostnames)
    if not isinstance(hostnames, list):
        raise Exception(f"name {hostnames} is not convertible to list")

    # at this point, hostnames is a list
    hostnames = [h.strip() for h in hostnames]

    if len(hostnames) > 1 and count > 1:
        _msg = (
            "If you set count>1, you should only specify one hostname with the %d formatter, not a list of hostnames."
        )
        raise Exception(_msg)

    if len(hostnames) == 1 and count > 0:
        hostname_spec = hostnames[0]
        count_range = range(count_offset, count_offset + count)
        if re.search(r"%\d{0,2}d", hostname_spec):
            hostnames = [hostname_spec % i for i in count_range]
        elif count > 1:
            hostnames = [f"{hostname_spec}{i:02}" for i in count_range]

    for hn in hostnames:
        if not is_valid_hostname(hn):
            raise Exception(f"Hostname '{hn}' does not seem to be valid")

    if len(hostnames) > MAX_DEVICES:
        raise Exception(f"You specified too many hostnames, max is {MAX_DEVICES}")
    return hostnames


def get_device_id_list(module):
    device_ids = module.params.get("device_ids")

    if isinstance(device_ids, str):
        device_ids = listify_string_name_or_id(device_ids)

    device_ids = [di.strip() for di in device_ids]

    for di in device_ids:
        if not is_valid_uuid(di):
            raise Exception(f"Device ID '{di}' does not seem to be valid")

    if len(device_ids) > MAX_DEVICES:
        raise Exception(f"You specified too many devices, max is {MAX_DEVICES}")
    return device_ids


def create_single_device(module, packet_conn, hostname):
    for param in ("hostnames", "operating_system", "plan"):
        if not module.params.get(param):
            raise Exception(f"{param} parameter is required for new device.")
    project_id = module.params.get("project_id")
    plan = module.params.get("plan")
    tags = module.params.get("tags")
    user_data = module.params.get("user_data")
    facility = module.params.get("facility")
    operating_system = module.params.get("operating_system")
    locked = module.params.get("locked")
    ipxe_script_url = module.params.get("ipxe_script_url")
    always_pxe = module.params.get("always_pxe")
    if operating_system != "custom_ipxe":
        for param in ("ipxe_script_url", "always_pxe"):
            if module.params.get(param):
                raise Exception(f"{param} parameter is not valid for non custom_ipxe operating_system.")

    device = packet_conn.create_device(
        project_id=project_id,
        hostname=hostname,
        tags=tags,
        plan=plan,
        facility=facility,
        operating_system=operating_system,
        userdata=user_data,
        locked=locked,
        ipxe_script_url=ipxe_script_url,
        always_pxe=always_pxe,
    )
    return device


def refresh_device_list(module, packet_conn, devices):
    device_ids = [d.id for d in devices]
    new_device_list = get_existing_devices(module, packet_conn)
    return [d for d in new_device_list if d.id in device_ids]


def wait_for_devices_active(module, packet_conn, watched_devices):
    wait_timeout = module.params.get("wait_timeout")
    wait_timeout = time.time() + wait_timeout
    refreshed = watched_devices
    while wait_timeout > time.time():
        refreshed = refresh_device_list(module, packet_conn, watched_devices)
        if all(d.state == "active" for d in refreshed):
            return refreshed
        time.sleep(5)
    timed_out_devices = [d.hostname for d in refreshed if d.state != "active"]
    raise Exception(f'Waiting for state "active" timed out for devices: {timed_out_devices}')


def wait_for_public_IPv(module, packet_conn, created_devices):
    def has_public_ip(addr_list, ip_v):
        return any(a["public"] and a["address_family"] == ip_v and a["address"] for a in addr_list)

    def all_have_public_ip(ds, ip_v):
        return all(has_public_ip(d.ip_addresses, ip_v) for d in ds)

    address_family = module.params.get("wait_for_public_IPv")

    wait_timeout = module.params.get("wait_timeout")
    wait_timeout = time.time() + wait_timeout
    while wait_timeout > time.time():
        refreshed = refresh_device_list(module, packet_conn, created_devices)
        if all_have_public_ip(refreshed, address_family):
            return refreshed
        time.sleep(5)

    raise Exception(
        f"Waiting for IPv{address_family} address timed out. Hostnames: {[d.hostname for d in created_devices]}"
    )


def get_existing_devices(module, packet_conn):
    project_id = module.params.get("project_id")
    return packet_conn.list_devices(project_id, params={"per_page": MAX_DEVICES})


def get_specified_device_identifiers(module):
    if module.params.get("device_ids"):
        device_id_list = get_device_id_list(module)
        return {"ids": device_id_list, "hostnames": []}
    elif module.params.get("hostnames"):
        hostname_list = get_hostname_list(module)
        return {"hostnames": hostname_list, "ids": []}


def act_on_devices(module, packet_conn, target_state):
    specified_identifiers = get_specified_device_identifiers(module)
    existing_devices = get_existing_devices(module, packet_conn)
    changed = False
    create_hostnames = []
    if target_state in ["present", "active", "rebooted"]:
        # states where we might create non-existing specified devices
        existing_devices_names = [ed.hostname for ed in existing_devices]
        create_hostnames = [hn for hn in specified_identifiers["hostnames"] if hn not in existing_devices_names]

    process_devices = [
        d
        for d in existing_devices
        if (d.id in specified_identifiers["ids"]) or (d.hostname in specified_identifiers["hostnames"])
    ]

    if target_state != "present":
        _absent_state_map = {}
        for s in PACKET_DEVICE_STATES:
            _absent_state_map[s] = packet.Device.delete

        state_map = {
            "absent": _absent_state_map,
            "active": {"inactive": packet.Device.power_on, "provisioning": None, "rebooting": None},
            "inactive": {"active": packet.Device.power_off},
            "rebooted": {
                "active": packet.Device.reboot,
                "inactive": packet.Device.power_on,
                "provisioning": None,
                "rebooting": None,
            },
        }

        # First do non-creation actions, it might be faster
        for d in process_devices:
            if d.state == target_state:
                continue
            if d.state in state_map[target_state]:
                api_operation = state_map[target_state].get(d.state)
                if api_operation is not None:
                    api_operation(d)
                    changed = True
            else:
                _msg = f"I don't know how to process existing device {d.hostname} from state {d.state} to state {target_state}"
                raise Exception(_msg)

    # At last create missing devices
    created_devices = []
    if create_hostnames:
        created_devices = [create_single_device(module, packet_conn, n) for n in create_hostnames]
        if module.params.get("wait_for_public_IPv"):
            created_devices = wait_for_public_IPv(module, packet_conn, created_devices)
        changed = True

    processed_devices = created_devices + process_devices
    if target_state == "active":
        processed_devices = wait_for_devices_active(module, packet_conn, processed_devices)

    return {"changed": changed, "devices": [serialize_device(d) for d in processed_devices]}


def main():
    module = AnsibleModule(
        argument_spec=dict(
            auth_token=dict(default=os.environ.get(PACKET_API_TOKEN_ENV_VAR), no_log=True),
            count=dict(type="int", default=1),
            count_offset=dict(type="int", default=1),
            device_ids=dict(type="list", elements="str"),
            facility=dict(),
            features=dict(type="dict"),
            hostnames=dict(type="list", elements="str", aliases=["name"]),
            tags=dict(type="list", elements="str"),
            locked=dict(type="bool", default=False, aliases=["lock"]),
            operating_system=dict(),
            plan=dict(),
            project_id=dict(required=True),
            state=dict(choices=ALLOWED_STATES, default="present"),
            user_data=dict(),
            wait_for_public_IPv=dict(type="int", choices=[4, 6]),
            wait_timeout=dict(type="int", default=900),
            ipxe_script_url=dict(default=""),
            always_pxe=dict(type="bool", default=False),
        ),
        required_one_of=[
            (
                "device_ids",
                "hostnames",
            )
        ],
        mutually_exclusive=[
            ("hostnames", "device_ids"),
            ("count", "device_ids"),
            ("count_offset", "device_ids"),
        ],
    )

    if not HAS_PACKET_SDK:
        module.fail_json(msg="packet required for this module")

    if not module.params.get("auth_token"):
        _fail_msg = f"if Packet API token is not in environment variable {PACKET_API_TOKEN_ENV_VAR}, the auth_token parameter is required"
        module.fail_json(msg=_fail_msg)

    auth_token = module.params.get("auth_token")

    packet_conn = packet.Manager(auth_token=auth_token)

    state = module.params.get("state")

    try:
        module.exit_json(**act_on_devices(module, packet_conn, state))
    except Exception as e:
        module.fail_json(msg=f"failed to set device state {state}, error: {e}", exception=traceback.format_exc())


if __name__ == "__main__":
    main()
